# Tutorial

This tutorial describes how to integrate new masking gadgets into [InspectorGadget]().

We illustrate this process with the Boolean Refresh gadget presented in [BartheBDFGSZ16]() (Algorithm 4b).

[Creating a new gadget](#creating)

[Adding our new gadget to InspectorGadget](#adding)

[Estimating our new gadget](#estimating)

***

<h2 id="creating">Creating a new gadget</h2>

### Gadget group

The first step in integrating a gadget into our tool is to determine its group ID or group name: a string formed by the concatenation of the gadget functionality and the types of [input and output masking](#io_masking_types).

In our example, the group name should be: `RefreshS1S1`.

If this group already exists, a file named `RefreshS1S1.py` should appear in `~/inspectorGadget/gogogadgets/gadgets/`. 

If not, create a file named `RefreshS1S1.py` in the above directory and add the following lines:
	
```
from ..atomic_operations import *
from ..parameters import * 
from ..import gadgets_dict
from ..gadgets_dict import sharing_type, security_level
```	

We will also need to import all the estimation functions linked to the gadgets in this new group into inspectorGadget. 

In the same directory, add the following line to the file `__init__.py`

```
from . import RefreshS1S1
```

**Important:** the order of the import lines in `__init__.py` is critical.
Each gadget group must be imported before any other gadget that needs it.
> For example, gadgets in the RefreshS1S1 group 
must be imported before those in the SecAddModpS1S1 group, as the gadgets in this group require the use of sub-gadgets 
belonging to the RefreshS1S1 group.

<h4 id="io_masking_types">Input and output masking types</h4>

In the current version, only Boolean (S1) and arithmetic (S2) masking types are integrated into [inspectorGadget](). If necessary, additional masking types can be added by updating the following line in `~/inspectorGadget/gogogadgets/gadget_class.py`:

```
sharing_type = Enum("sharing_type", ["Boolean", "Arithmetic"])
```


### Estimation functions

Each gadget in the gadget group must define two estimation functions:

```
def <group_name>_<gadget_name>_cost(mask_params) :
	...
```

and

```
def <group_name>_<gadget_name>_memory(mask params) :
	...
```

for the performance and memory footprint estimation functions. 

`<gadget_name>` can be any string, e.g. the reference of the paper into which it was introduced.

In our example, we thus have two functions named: `RefreshS1S1_BartheBDFGSZ16_4b_cost` and `RefreshS1S1_BartheBDFGSZ16_4b_memory`.


For each function, we need to recover the cost of each atomic operations used by the gadget. These atomic operations are identified by an alias listed
in the `set_atomic_operations_costs` function in `~/inspectorGadget/gogogadgets/atomic_operations.py`. 
They can be retrieved using calls of the form: `atomic_operations.get("<atomic_operation>").get("cost")`.
    
For example, for our RefreshS1S1 gadget, we retrieve the atomic cost of the xor operation with:

```
xor_op = atomic_operations.get("xor_op").get("cost")
```

**Note:** Some gadgets require the use of sub-gadgets. To retrieve the performance estimation/memory footprint estimations of 
these sub-gadgets, we use the following call structure: 
	
```
gadgets_dict.{get_cost|get_memory}( \
	mask_params, \
 	"<gadget_group>", \
 	sharing_type.{Boolean|Arithmetic}, \
 	sharing_type.{Boolean|Arithmetic}, \
 	name = None|"<gadget_name>" \
 	)
``` 		

#### Performance estimation function

All the details concerning the definition of performance estimation functions are described in Appendix A.3 of our paper. The function should return the sum of all atomic operations performed in the implementation used as a reference for the gadget.

#### Memory estimation function

For the memory footprint estimation function, we refer to Appendix A.2 of our paper for a detailed example of its construction, particularly in the case of a more complex gadget requiring the use of sub-gadgets. 



<h2 id="adding">Adding our new gadget to InspectorGadget</h2>

To add our gadget to [inspectorGadget](), simply use the command described at the end of the `gogogadgets` section of `README.md`.

For our example, we need to add the following command after our two estimation functions:

```
gadgets_dict.add_new_gadget( \
	"Refresh", \
	sharing_type.Boolean, \
	sharing_type.Boolean, \
	"BartheBDFGSZ16_4b", \	
	RefreshS1S1_BartheBDFGSZ16_4b_cost, \
	RefreshS1S1_BartheBDFGSZ16_4b_memory, \
	"https://eprint.iacr.org/2015/506.pdf", \
	security_lvl = security_level.SNI \
	)
```	

**Note:** Our RefreshS1S1 gadget has been proven SNI, so we can set its security level to SNI. If the security level is not proven, set `security_lvl` to `None`.


<h2 id="estimating">Estimating our new gadget</h2>

To estimate a gadget, follow the steps from `README.md`.

For example. in order to estimate the costs of our RefreshS1S1 gadget with 5 shares on a Cortex M3 microcontroler, with a random generation cost of 32 CCE, we run:

```
python3 inspector.py \
	--component_config_file "examples_hardware_config Cortex_M3_rand=32.toml" \
	--params_config_file    "examples_masking_config/kyber_params.toml" \
	--gadget_func           "Refresh" \
	--input_type            "Boolean" \
	--output_type           "Boolean" \
	--gadget_name           "BartheBDFGSZ16_4b" \
	--list_n                "[5]"
```














