# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import tomlkit

"""
	File that contains the atomic operations cost dictionary atomic_operations. 
	The dictionary contains two types of operations:
	- assembler instructions
	- high level operations
	Assembler instructions are used to define the cost of high-level operations. 
	The latter are those which will be used in the functions for estimating the 
	cost in CCE of the various gadgets. All costs are defined in the hardware
	configuration file.
	Memory consumption is not considered for atomic operations so this parameter 
	is set to 0 for all operations.
"""

list_atomic_operations    = ['LDR', 'STR', 'MOV', 'ADD', 'SUB', 'MULT', 'EORS', 'ORRS', 'BIC', 'AND', 'LSL', 'LSR', 'ROR', 'MVN', 'UDIV','rand_generation','rand_generation_prime_field', 'modulo_Q_op']
list_hardware_parameters  = list_atomic_operations + ['register_space','memory_space']

def fill_with_null():
	global atomic_operations

	for op in list_atomic_operations:
		atomic_operations[op] = {'cost': 0}

	# High-level operations whose cost depends on assembly instructions
	atomic_operations["store_op"]      = {"cost": atomic_operations["STR"]["cost"]}
	atomic_operations["load_op"]       = {"cost": atomic_operations["LDR"]["cost"]}
	atomic_operations["mov_op"]        = {"cost": atomic_operations["MOV"]["cost"]}
	atomic_operations["not_op"]        = {"cost": atomic_operations["MVN"]["cost"]}
	atomic_operations["mask_op"]       = {"cost": atomic_operations["LSL"]["cost"]}
	atomic_operations["leftshift_op"]  = {"cost": atomic_operations["LSL"]["cost"]}
	atomic_operations["rightshift_op"] = {"cost": atomic_operations["LSR"]["cost"]}
	atomic_operations["ror_op"]        = {"cost": atomic_operations["ROR"]["cost"]}
	atomic_operations["xor_op"]        = {"cost": atomic_operations["EORS"]["cost"]}
	atomic_operations["or_op"]         = {"cost": atomic_operations["ORRS"]["cost"]}
	atomic_operations["and_op"]        = {"cost": atomic_operations["AND"]["cost"]}
	atomic_operations["add_op"]        = {"cost": atomic_operations["ADD"]["cost"]}
	atomic_operations["sub_op"]        = {"cost": atomic_operations["SUB"]["cost"]}
	atomic_operations["mult_op"]       = {"cost": atomic_operations["MULT"]["cost"]}
	atomic_operations["udiv_op"]       = {"cost": atomic_operations["UDIV"]["cost"]}

	# Memory cost of all operations set to 0
	for atomic_op in atomic_operations.keys():
		atomic_operations[atomic_op]["memory"] = 0

hardware_model            = None
config_hardware_file_name = None

# Atomic operations dictionary definition
atomic_operations = dict()
fill_with_null()


def set_atomic_operations_costs(config_component_parameters_file):
	global hardware_model, atomic_operations, config_hardware_file_name

	try:
		with open(config_component_parameters_file, mode="rt", encoding="utf-8") as fp:
			hardware_model = tomlkit.load(fp)

		# verify that all necessary parameters are set
		for v in list_hardware_parameters:
			_ = hardware_model[v]

		config_hardware_file_name = config_component_parameters_file

	except Exception as e:
		print(e)
		print(f'Error [atomic_operations.set_atomic_operations_costs]: config file does not set all instructions correctly... aborting.')
		hardware_model = None
		config_hardware_file_name = None
		fill_with_null()
		exit(1)


	# Assembly instructions
	for op in list_atomic_operations:
		atomic_operations[op] = {'cost': hardware_model[op]}

	# High-level operations whose cost depends on assembly instructions
	atomic_operations["store_op"]      = {"cost": atomic_operations["STR"]["cost"]}
	atomic_operations["load_op"]       = {"cost": atomic_operations["LDR"]["cost"]}
	atomic_operations["mov_op"]        = {"cost": atomic_operations["MOV"]["cost"]}
	atomic_operations["not_op"]        = {"cost": atomic_operations["MVN"]["cost"]}
	atomic_operations["mask_op"]       = {"cost": atomic_operations["LSL"]["cost"]}
	atomic_operations["leftshift_op"]  = {"cost": atomic_operations["LSL"]["cost"]}
	atomic_operations["rightshift_op"] = {"cost": atomic_operations["LSR"]["cost"]}
	atomic_operations["ror_op"]        = {"cost": atomic_operations["ROR"]["cost"]}
	atomic_operations["xor_op"]        = {"cost": atomic_operations["EORS"]["cost"]}
	atomic_operations["or_op"]         = {"cost": atomic_operations["ORRS"]["cost"]}
	atomic_operations["and_op"]        = {"cost": atomic_operations["AND"]["cost"]}
	atomic_operations["add_op"]        = {"cost": atomic_operations["ADD"]["cost"]}
	atomic_operations["sub_op"]        = {"cost": atomic_operations["SUB"]["cost"]}
	atomic_operations["mult_op"]       = {"cost": atomic_operations["MULT"]["cost"]}
	atomic_operations["udiv_op"]       = {"cost": atomic_operations["UDIV"]["cost"]}

	# Memory cost of all operations set to 0
	for atomic_op in atomic_operations.keys():
		atomic_operations[atomic_op]["memory"] = 0


