# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from enum import Enum
import numpy as np

# Type of masking possible on entry and exit of a gadget
sharing_type = Enum("sharing_type", ["Boolean", "Arithmetic"])
# Type of security level of a gadget
security_level = Enum("security_level", ["NI", "SNI", "PINI", "S_PI_NI"])

def get_group(gadget_key):
	gadget_key_split = gadget_key.split('_')
	if len(gadget_key_split) == 1:
		# gadget_key is actually a group key
		return gadget_key, None

	group = gadget_key_split[0]
	name  = gadget_key[len(group)+1:]

	return group, name

def get_gadget_info(gadget_key):
	"""Gives information about a gadget that can be obtained from its key

	Breaks down the given input gadget's key to deduce its functionality, the 
	type of sharing it takes in input and output, and the gadget name.

	Input:
		gadget_key (string): Key of a gadget or a group in _gadgets_set

	Output:
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group
	"""
	group, name  = get_group(gadget_key)
	gadget_group = group.split('S')

	try:
		gadget_func      = group[:-(len(gadget_group[-2])+len(gadget_group[-1])+2)]
		input_type       = sharing_type(int(gadget_group[-2]))
		output_type      = sharing_type(int(gadget_group[-1]))
	except:
		print(f'Error [gadget_class.get_gadget_info]: {gadget_key=} is not formated correctly... aborting')
		exit(1)

	return gadget_func, input_type, output_type, name

def get_group_key(gadget_func, input_type, output_type):
	group = gadget_func+"S"+str(input_type.value)+"S"+str(output_type.value)
	return group.upper()

def get_key(gadget_func, input_type, output_type, name):
	group = get_group_key(gadget_func, input_type, output_type)
	if name is None:
		return group

	return (group+"_"+name).upper()

def get_group_and_key(gadget_func, input_type, output_type, name):
	group = get_group_key(gadget_func, input_type, output_type)
	if name is None:
		return group, group

	return group, (group+"_"+name).upper()

class group_c():
	def __init__(self, name):
		# create a group from a group name
		self.name           = name
		self.gadgets        = {}
		self.default_gadget = None

	def __init__(self, gadget):
		# create a group from a gadget
		self.name           = gadget.group
		self.gadgets        = {gadget.name: gadget}
		self.default_gadget = None

	def add_new_gadget(self, gadget):
		try:
			_ = self.gadgets[gadget.name]
			print(f"Warning [group_c.add_new_gadget]: Gadget with name {gadget.name} already exist in group {self.name}.")
		except:
			self.gadgets[gadget.name] = gadget

	def remove_gadget(self, gadget_name):
		try:
			del self.gadgets[gadget_name.upper()]
		except:
			print(f"Warning [group_c.remove_gadget]: Gadget with name {gadget_name} was not present in group {self.name}.")

	def gadgetExists(self, gadget_name):
		return gadget_name.upper() in self.gadgets.keys()

	def get_gadgets_name(self):
		return self.gadgets.keys()

	def get_gadget(self, gadget_name = None):
		if len(self.gadgets.keys()) == 0:
			return None

		if gadget_name is None:
			return None

		if gadget_name == 'ANY':
			list_gadgets = list(self.gadgets.keys())
			idx = np.random.randint(0,len(list_gadgets))
			return self.gadgets[list_gadgets[idx]]

		if self.gadgetExists(gadget_name):
			return self.gadgets[gadget_name.upper()]
		elif self.gadgetExists(gadget_name[len(self.name)+1:]):
			return self.gadgets[gadget_name[len(self.name)+1:].upper()]
		else:
			print(f"Warning [group_c.get_gadget]: Gadget with name {gadget_name} is not in group {self.name}.")
			return None

	def __str__(self):

		out  = f'Group {self.name}:\n'
		for g in self.gadgets.keys():
			out += f'{self.gadgets[g]}\n'

		return out


class gadget_c():

	def __init__(self, gadget_func, input_type, output_type, name, cost_estimation, memory_estimation, ref, security_lvl = None):
		
		try:
			self.func         = gadget_func.upper()
			self.input_type   = input_type
			self.output_type  = output_type
			self.name         = name.upper()
			self.cost_func    = cost_estimation
			self.memo_func    = memory_estimation
			self.ref          = ref
			self.security_lvl = security_lvl
		except Exception as e:
			print(e)
			print('Error [gadget_c.__init__()]: inputs are not correct... aborting')
			exit()

		self.key         = get_key(gadget_func, input_type, output_type, name)
		self.group       = get_group_key(gadget_func, input_type, output_type)

		self.sub_gadgets = []
		self.sub_groups  = []
		
	def __str__(self):
		out = f'\t {self.name}\n'
		out += f'\t\t {self.func}\n'
		out += f'\t\t {self.input_type}\n'
		out += f'\t\t {self.output_type}\n'
		out += f'\t\t {self.ref}\n'
		out += f'\t\t {self.security_lvl}\n'
		out += f'\t\t sub-groups:\n'
		for v in self.sub_groups:
			out += f'\t\t\t {v}\n'
		out += f'\t\t sub-gadgets:\n'
		for v in self.sub_gadgets:
			out += f'\t\t\t {v}\n'

		return out
		




