# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import ceil, log, floor

from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def BsSecA2BPowerOf2S2S1_BronchainC22_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the bitsliced 
    SecA2BPowerOf2 algorithm of [BC22]

    Gives an estimate of the number of CPU cycles of the bitsliced 
    SecA2BPowerOf2 algorithm of [BC22], according to the parameters given at the 
    input of the function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void seca2b(size_t nshares, size_t kbits, uint32_t *in, size_t in_msk_stride,
                          size_t in_data_stride)
    """
    
    n = mask_params.n
    if(n == 1):
        return 0
    else:
        bitsize_shares = mask_params.bitsize_shares
        sub_op = atomic_operations.get("sub_op").get("cost")
        rightshift_op = atomic_operations.get("rightshift_op").get("cost")
        BsCopySharing_cost = gadgets_dict.get_cost(mask_params, "BsCopySharing", sharing_type.Boolean, sharing_type.Boolean, name = None)
        BsSecAdd_cost = gadgets_dict.get_cost(mask_params, "BsSecAdd", sharing_type.Boolean, sharing_type.Boolean, name = None)
        # Computation of floor(n/2) and n-floor(n/2)
        cost = sub_op+rightshift_op
        # BsSecA2B of half the shares
        scheme_parameters_copy = deepcopy(mask_params)
        half_order = floor(n/2)
        scheme_parameters_copy.n = half_order
        cost += BsSecA2BPowerOf2S2S1_BronchainC22_cost(scheme_parameters_copy)
        scheme_parameters_copy2 = deepcopy(mask_params)
        scheme_parameters_copy2.n = n-half_order
        cost += BsSecA2BPowerOf2S2S1_BronchainC22_cost(scheme_parameters_copy2)
        cost += bitsize_shares*BsCopySharing_cost
        cost += BsSecAdd_cost
    return cost

def BsSecA2BPowerOf2S2S1_BronchainC22_memory(mask_params):
    """
    Estimation of the maximum of memory used in the bitsliced SecA2BPowerOf2 
    algorithm of [BC22]

    Gives an estimate of the maximum of memory used by the bitsliced 
    SecA2BPowerOf2 algorithm of [BC22], according to the parameters given at the 
    input of the function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void seca2b(size_t nshares, size_t kbits, uint32_t *in, size_t in_msk_stride,
                          size_t in_data_stride)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares

    if(n == 1):
        memory = 0
    else:
        # Computation of the memory taken by the variables
        half_order = n//2
        # size_t: nshares_low, nshares_high
        memory = ceil(log(n, 2))
        # Masked: expanded_low, expanded_high
        memory += 2*n*bitsize_shares*32

        # recurcive calls
        scheme_parameters_copy = deepcopy(mask_params)
        scheme_parameters_copy.n = half_order
        memory_1 = BsSecA2BPowerOf2S2S1_BronchainC22_memory(scheme_parameters_copy)
        scheme_parameters_copy2 = deepcopy(mask_params)
        scheme_parameters_copy2.n = n-half_order
        memory_2 = BsSecA2BPowerOf2S2S1_BronchainC22_memory(scheme_parameters_copy2)

        
        BsSecAdd_memory = gadgets_dict.get_memory(mask_params, "BsSecAdd", sharing_type.Boolean, sharing_type.Boolean, name = None)
        
        memory += max(memory_1, memory_2, BsSecAdd_memory)

    return memory
    
gadgets_dict.add_new_gadget("BsSecA2BPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, "BronchainC22", BsSecA2BPowerOf2S2S1_BronchainC22_cost, BsSecA2BPowerOf2S2S1_BronchainC22_memory, "https://eprint.iacr.org/2022/158.pdf", security_lvl = security_level.PINI)

