# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import ceil, log
from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def BsSecAddModpS1S1_BronchainC22_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the bitsliced SecAddModp 
    algorithm of [BC22]

    Gives an estimate of the number of CPU cycles of the bitsliced SecAddModp 
    algorithm of [BC22], according to the parameters given at the input of the 
    function. The number of bits used as input to the function is taken into 
    account in the calculations.

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void secadd_modp(size_t nshares, size_t kbits, uint32_t q, uint32_t *out,
                               size_t out_msk_stride, size_t out_data_stride,
                               const uint32_t *in1, size_t in1_msk_stride,
                               size_t in1_data_stride, const uint32_t *in2,
                               size_t in2_msk_stride, size_t in2_data_stride)
    """
    

    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    mov_op = (atomic_operations.get("mov_op")).get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    add_op = (atomic_operations.get("add_op")).get("cost")
    sub_op = (atomic_operations.get("sub_op")).get("cost")
    BsSecAdd_cost = gadgets_dict.get_cost(scheme_parameters_adder,"BsSecAdd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    cost = BsSecAdd_cost
    cost += leftshift_op+add_op+sub_op
    cost += BsSecAdd_cost
    cost += n*mov_op
    cost += n*(bitsize_shares-1)*mov_op
    cost += BsSecAdd_cost
    return cost

def BsSecAddModpS1S1_BronchainC22_memory(mask_params):
    """
    Estimation of the maximum of memory used in the bitsliced SecAddModp 
    algorithm of [BC22]

    Gives an estimate of the maximum of memory used by the bitsliced SecAddModp 
    algorithm of [BC22], according to the parameters given at the input of the 
    function. The number of bits used as input to the function is taken into 
    account in the calculations.

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void secadd_modp(size_t nshares, size_t kbits, uint32_t q, uint32_t *out,
                               size_t out_msk_stride, size_t out_data_stride,
                               const uint32_t *in1, size_t in1_msk_stride,
                               size_t in1_data_stride, const uint32_t *in2,
                               size_t in2_msk_stride, size_t in2_data_stride)
    """
    global scheme_related_parameters
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    scheme_parameters_adder = deepcopy(mask_params)
    scheme_parameters_adder.set_bitsize_shares(1)
    BsSecAdd_memory = gadgets_dict.get_memory(scheme_parameters_adder,"BsSecAdd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    # Computation of the space taken by the variables
    # uint32_t p[(kbits + 1) * nshares], s[(kbits + 1) * nshares], sp[(kbits + 1) * nshares]
    memory = 3*(bitsize_shares+1)*n
    # uint32_t b[nshares]
    memory += n
    # uint32_t p[kbits * nshares]
    memory += bitsize_shares*n
    memory += BsSecAdd_memory
    return memory

gadgets_dict.add_new_gadget("BsSecAddModp", sharing_type.Boolean, sharing_type.Boolean, "BronchainC22", BsSecAddModpS1S1_BronchainC22_cost, BsSecAddModpS1S1_BronchainC22_memory, "https://eprint.iacr.org/2022/158.pdf", security_lvl = security_level.PINI)

