# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def PiniAndCoreS1S1_CassiersS20_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the PINI And Core 
    algorithm of [CS20]

    Gives an estimate of the number of CPU cycles of the PINI And Core algorithm 
    of [CS20], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Trivially and Efficiently Composing Masked Gadgets with Probe 
    Isolating Non-Interference
    Link: https://eprint.iacr.org/2018/438.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: uint32_t pini_and_core(uint32_t a, uint32_t b, uint32_t r)
    """
    
    eors_assembler_op = atomic_operations.get("EORS").get("cost")
    and_assembleur_op = atomic_operations.get("AND").get("cost")
    bic_assembler_op = atomic_operations.get("BIC").get("cost")
    cost = 2*eors_assembler_op+and_assembleur_op+bic_assembler_op
    return cost

def PiniAndCoreS1S1_CassiersS20_memory(mask_params):
    """
    Estimation of the maximum of memory used in the PINI And Core algorithm 
    of [CS20]

    Gives an estimate of the maximum of memory used by the PINI And Core 
    algorithm of [CS20], according to the parameters given at the input of the 
    function. The number of bits used as input to the function is taken into 
    account in the calculations.

    Paper: Trivially and Efficiently Composing Masked Gadgets with Probe 
    Isolating Non-Interference
    Link: https://eprint.iacr.org/2018/438.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: uint32_t pini_and_core(uint32_t a, uint32_t b, uint32_t r)
    """
    # uint32_t: tmp, s
    memory = 2*32
    return memory

def BsSecAndS1S1_CassiersS20_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the bitsliced PINI SecAnd 
    algorithm of [CS20]

    Gives an estimate of the number of CPU cycles of the bitsliced PINI SecAnd 
    algorithm of [CS20], according to the parameters given at the input of the 
    function. The number of bits used as input to the function is taken into 
    account in the calculations.

    Paper: Trivially and Efficiently Composing Masked Gadgets with Probe 
    Isolating Non-Interference
    Link: https://eprint.iacr.org/2018/438.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void masked_and_c(size_t nshares, uint32_t *z, size_t z_stride,
                                const uint32_t *a, size_t a_stride, const uint32_t *b,
                                size_t b_stride)
    """
    
    n = mask_params.n
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    mov_op = atomic_operations.get("mov_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    mask_op = atomic_operations.get("mask_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    PiniAndCore_cost = PiniAndCoreS1S1_CassiersS20_cost(mask_params)
    # Loop to compute ztmp_i = a_i & b_i
    cost = n*and_op
    # And core
    for i in range(n-1):
        for j in range(i+1, n):
            cost += rand_generation
            cost += xor_op+PiniAndCore_cost
            cost += xor_op+PiniAndCore_cost
    # Loop to compute z_i = ztmp_i
    cost += n*mov_op
    return cost

def BsSecAndS1S1_CassiersS20_memory(mask_params):
    """
    Estimation of the maximum of memory used in the PINI SecAnd algorithm of 
    [CS20]

    Gives an estimate of the maximum of memory used by the PINI SecAnd algorithm 
    of [CS20], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Trivially and Efficiently Composing Masked Gadgets with Probe 
    Isolating Non-Interference
    Link: https://eprint.iacr.org/2018/438.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void masked_and_c(size_t nshares, uint32_t *z, size_t z_stride,
                                const uint32_t *a, size_t a_stride, const uint32_t *b,
                                size_t b_stride)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    PiniAndCore_mem = PiniAndCoreS1S1_CassiersS20_memory(mask_params)
    # Masked uint32_t: ztmp
    memory = n*32
    # uint32_t: r
    memory += 32
    # PINI mem
    memory += PiniAndCore_mem
    return memory

gadgets_dict.add_new_gadget("BsSecAnd", sharing_type.Boolean, sharing_type.Boolean, "CassiersS20", BsSecAndS1S1_CassiersS20_cost, BsSecAndS1S1_CassiersS20_memory, "https://eprint.iacr.org/2018/438.pdf", security_lvl = security_level.PINI)