# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import log, ceil

from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def Dense2bitsliced32bitsS2S2_BronchainC22_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Dense2bitsliced32bits 
    algorithm of [BC22] 

    Gives an estimate of the number of CPU cycles of the Dense2bitsliced32bits 
    algorithm of [BC22], according to the parameters given at the input of the 
    function.

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit 
           with Application to Lattice-Based KEMs
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/masked_representations.c
    Function: void masked_dense2bitslice_opt_u32(size_t nshares, size_t coeffs_size, uint32_t *bitslice,
                                                 size_t bitslice_msk_stride, size_t bitslice_data_stride,
                                                 const uint32_t *dense, size_t dense_msk_stride, size_t dense_data_stride)
    """
    
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    mov_op = atomic_operations.get("mov_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    Transpose32S2S2_cost = gadgets_dict.get_cost(mask_params, "Transpose32", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    cost = 0
    # Computing a[i]
    for d in range(n):
        for i in range(32):
            cost += mov_op
        cost += Transpose32S2S2_cost
        for i in range(bitsize_shares):
            cost += mov_op
    return cost

def Dense2bitsliced32bitsS2S2_BronchainC22_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Dense2bitsliced32bits 
    algorithm of [BC22]

    Gives an estimate of the maximum of memory used by the Dense2bitsliced32bits 
    algorithm of [BC22], according to the parameters given at the input of 
    the function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit 
           with Application to Lattice-Based KEMs
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/masked_representations.c
    Function: void masked_dense2bitslice_opt_u32(size_t nshares, size_t coeffs_size, uint32_t *bitslice,
                                                 size_t bitslice_msk_stride, size_t bitslice_data_stride,
                                                 const uint32_t *dense, size_t dense_msk_stride, size_t dense_data_stride)
    
    """
    bitsize_shares = mask_params.bitsize_shares
    Transpose32S2S2_mem = gadgets_dict.get_memory(mask_params, "Transpose32", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    # uint32_t: a[32]
    memory = 32*bitsize_shares
    # Transpose
    memory += Transpose32S2S2_mem
    return memory

gadgets_dict.add_new_gadget("Dense2bitsliced32bits", sharing_type.Arithmetic, sharing_type.Arithmetic, "BronchainC22", Dense2bitsliced32bitsS2S2_BronchainC22_cost, Dense2bitsliced32bitsS2S2_BronchainC22_memory, "https://eprint.iacr.org/2022/158.pdf", security_lvl = None)

