# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def ExpandS1S1_CoronGV14_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Expand algorithm of 
    [CGV14] (Algorithm 5)

    Gives an estimate of the number of CPU cycles of the Expand algorithm of 
    [CGV14], according to the parameters given at the input of the function. 

    Paper: Secure Conversion Between Boolean and Arithmetic Masking of Any Order
    Link: http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    """
    
    n = mask_params.n
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    
    cost = 0
    for i in range(n):
        # Generating random
        cost += rand_generation+and_op
        # y_{2i}
        cost += xor_op
    return cost

def ExpandS1S1_CoronGV14_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Expand algorithm of 
    [CGV14] (Algorithm 5)

    Gives an estimate of the maximum of memory used by the Expand algorithm of 
    [CGV14], according to the parameters given at the input of the function.

    Paper: Secure Conversion Between Boolean and Arithmetic Masking of Any Order
    Link: http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    """
    # uint32_t: r_i
    memory = mask_params.bitsize_shares
    return memory

gadgets_dict.add_new_gadget("Expand", sharing_type.Boolean, sharing_type.Boolean, "CoronGV14", ExpandS1S1_CoronGV14_cost, ExpandS1S1_CoronGV14_memory, "http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf", security_lvl = None)

