# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

additional_KyberCompress_params = dict()
additional_KyberCompress_params["d"] = 1

def KyberCompress256coeffsS2S1_OderSPG18_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Compress algorithm of 
    [OSPG18] (Algorithm 2, MDecode)

    Gives an estimate of the number of CPU cycles of the Compress algorithm of 
    [OSPG18], according to the parameters given at the input of the function. 

    Paper: Practical CCA2-Secure and Masked Ring-LWE Implementation
    Link: https://tches.iacr.org/index.php/TCHES/article/view/836/788

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/crypto_kem/kyber768/m4/masked-poly.c
    Function: void masked_poly_tomsg(masked_u8_msgbytes* msg, masked_poly* a)
    """
    
    n = mask_params.n
    
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget KyberCompressS2S1_OderSPG18 only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None

    rand_generation = atomic_operations.get("rand_generation").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    load_op = atomic_operations.get("load_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    
    SecA2BModPowerOf2_cost = gadgets_dict.get_cost(mask_params, "SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    TransformPower2_cost   = gadgets_dict.get_cost(mask_params, "TransformPower2", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "OderSPG18")
    
    # Double loop to cycle through all coefficients to be compressed
    # a <- a-(q/4) % q
    cost = 256*(sub_op+add_op+modulo_Q_op)

    # in case using BeirendonckDV21_SingleLookUp, it must be initialized once
    try:
        assert(gadgets_dict._gadgets_set["SECA2BMODPOWEROF2S2S1"].default_gadget.name == "BeirendonckDV21_SingleLookUp".upper())
        SecA2BModPowerOf2Init_cost = gadgets_dict.get_cost(mask_params, "SecA2BModPowerOf2Init", sharing_type.Arithmetic, sharing_type.Boolean, name = "BeirendonckDV21_SingleLookUp")
        # Table initialization for A2B conversion
        cost += SecA2BModPowerOf2Init_cost
    except:
        pass

    cost += 256*TransformPower2_cost

    cost += 256*sub_op
    cost += 256*SecA2BModPowerOf2_cost
    cost += 256*2*(add_op+rightshift_op+and_op+leftshift_op)
    
    # output is over 16 bits
    mask_params.set_bitsize_shares(16)

    return cost

def KyberCompress256coeffsS2S1_OderSPG18_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Compress algorithm of 
    [OSPG18] (Algorithm 2, MDecode)

    Gives an estimate of the maximum of memory used by the Compress algorithm of 
    [OSPG18], according to the parameters given at the input of the function.

    Paper: Practical CCA2-Secure and Masked Ring-LWE Implementation
    Link: https://tches.iacr.org/index.php/TCHES/article/view/836/788

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/crypto_kem/kyber768/m4/masked-poly.c
    Function: void masked_poly_tomsg(masked_u8_msgbytes* msg, masked_poly* a)
    """
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget KyberCompressS2S1_OderSPG18 only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    # Computation of the memory taken by the variables
    bitsize_shares = mask_params.bitsize_shares
    # Shares: a_1, a_2
    memory = 2*bitsize_shares
    # uint16_t: c[0], c[1]
    memory += 2*16

    TransformPower2_memory = gadgets_dict.get_memory(mask_params, "TransformPower2", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "OderSPG18")
    SecA2BModPowerOf2_memory = gadgets_dict.get_memory(mask_params, "SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
        
    try:
        assert(gadgets_dict._gadgets_set["SECA2BMODPOWEROF2S2S1"].default_gadget.name == "BeirendonckDV21_SingleLookUp".upper())
        SecA2BModPowerOf2Init_memory = gadgets_dict.get_memory(mask_params, "SecA2BModPowerOf2Init", sharing_type.Arithmetic, sharing_type.Boolean, name = "BeirendonckDV21_SingleLookUp")
        # Table initialization for A2B conversion
        memory += max([SecA2BModPowerOf2_memory,SecA2BModPowerOf2Init_memory,TransformPower2_memory])
    except:
        memory += max(SecA2BModPowerOf2_memory,TransformPower2_memory)
   
    # output is over 16 bits
    mask_params.set_bitsize_shares(16)

    return memory

gadgets_dict.add_new_gadget("KyberCompress256coeffs", sharing_type.Arithmetic, sharing_type.Boolean, "OderSPG18", KyberCompress256coeffsS2S1_OderSPG18_cost, KyberCompress256coeffsS2S1_OderSPG18_memory, "https://tches.iacr.org/index.php/TCHES/article/view/836/788", security_lvl = None)

def KyberCompress256coeffsS2S1_BosGRSV21_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Compress algorithm of 
    [BGRSV21] (Algorithm 1)

    Gives an estimate of the number of CPU cycles of the Compress algorithm of 
    [BGRSV21], according to the parameters given at the input of the function. 

    Paper: Masking Kyber: First- and Higher-Order Implementations
    Link: https://eprint.iacr.org/2021/483.pdf
    """
    
    load_op = atomic_operations.get("load_op").get("cost")
    store_op = atomic_operations.get("store_op").get("cost")
    or_op = atomic_operations.get("or_op").get("cost")
    not_op = atomic_operations.get("not_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")

    bitsize_shares = mask_params.bitsize_shares
    n = mask_params.n

    # for the SecA2BModp, the share size must be incresed by 1 bit
    mask_params.set_bitsize_shares(bitsize_shares+1)
    SecA2BModp_cost = gadgets_dict.get_cost(mask_params,"SecA2BModp", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    mask_params.set_bitsize_shares(bitsize_shares)

    # We use the Dense2bitsliced implementation for 32bits as it is very efficient
    # it will have to be applied 8 times for 256 coefficients
    Dense2bitsliced32bits_cost = gadgets_dict.get_cost(mask_params,"Dense2bitsliced32bits", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    
    # line 1 to 3 of Algorithm 1
    cost = 256*(add_op+modulo_Q_op+SecA2BModp_cost)
    # Bitslicing (line 4)
    cost += 8*Dense2bitsliced32bits_cost

    # after bitslice shares size is set to 256
    mask_params.set_bitsize_shares(256)

    # Compression computation using the logic formula
    Refresh_cost = gadgets_dict.get_cost(mask_params,"Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecNot_cost = gadgets_dict.get_cost(mask_params,"SecNot", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecXOR_cost = gadgets_dict.get_cost(mask_params,"SecXOR", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAnd_cost = gadgets_dict.get_cost(mask_params,"SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    
    # line 5 to 10 of Algorithm 1
    cost += 2*SecNot_cost+2*Refresh_cost+4*SecAnd_cost+2*SecXOR_cost

    return cost

def KyberCompress256coeffsS2S1_BosGRSV21_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Compress algorithm of 
    [BGRSV21] (Algorithm 1)

    Gives an estimate of the maximum of memory used by the Compress algorithm of 
    [BGRSV21], according to the parameters given at the input of the function.

    Paper: Masking Kyber: First- and Higher-Order Implementations
    Link: https://eprint.iacr.org/2021/483.pdf
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares

    # this is to store the output of bitslicing
    memory = 256*n*bitsize_shares

    # for the SecA2BModp, the share size must be incresed by 1 bit
    mask_params.set_bitsize_shares(bitsize_shares+1)
    SecA2BModp_memory = gadgets_dict.get_memory(mask_params,"SecA2BModp", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    mask_params.set_bitsize_shares(bitsize_shares)

    # We use the Dense2bitsliced implementation for 32bits as it is very efficient
    Dense2bitsliced32bits_memory = gadgets_dict.get_memory(mask_params,"Dense2bitsliced32bits", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    
    # after bitslice shares size is set to 256
    mask_params.set_bitsize_shares(256)

    # Compression computation using the logic formula
    Refresh_memory = gadgets_dict.get_memory(mask_params,"Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAnd_memory = gadgets_dict.get_memory(mask_params,"SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecNot_memory = gadgets_dict.get_memory(mask_params,"SecNot", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecXOR_memory = gadgets_dict.get_memory(mask_params,"SecXOR", sharing_type.Boolean, sharing_type.Boolean, name = None)
    
    # max of the sequential operations
    memory += max([SecA2BModp_memory, Dense2bitsliced32bits_memory, Refresh_memory, SecAnd_memory, SecNot_memory, SecXOR_memory])
    
    return memory

gadgets_dict.add_new_gadget("KyberCompress256coeffs", sharing_type.Arithmetic, sharing_type.Boolean, "BosGRSV21", KyberCompress256coeffsS2S1_BosGRSV21_cost, KyberCompress256coeffsS2S1_BosGRSV21_memory, "https://eprint.iacr.org/2021/483.pdf", security_lvl = security_level.SNI)

def KyberCompress256coeffsS2S1_BronchainC22_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the bitsliced Compress 
    algorithm of [BC22]

    Gives an estimate of the number of CPU cycles of the bitsliced Compress 
    algorithm of [BC22], according to the parameters given at the input of the 
    function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    8 times the KyberCompress32coeffs function
    """
    
    KyberCompress32coeffs_cost = gadgets_dict.get_cost(mask_params, "KyberCompress32coeffs", sharing_type.Arithmetic, sharing_type.Boolean, name = "BronchainC22")
    cost = 8*KyberCompress32coeffs_cost
    return cost

def KyberCompress256coeffsS2S1_BronchainC22_memory(mask_params):
    """
    Estimation of the maximum of memory used in the bitsliced Compress 
    algorithm of [BC22]

    Gives an estimate of the maximum of memory used by the bitsliced Compress 
    algorithm of [BC22], according to the parameters given at the input of the 
    function.

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    8 times the KyberCompress32coeffs function
    """
    memory = gadgets_dict.get_memory(mask_params, "KyberCompress32coeffs", sharing_type.Arithmetic, sharing_type.Boolean, name = "BronchainC22")
    return memory

gadgets_dict.add_new_gadget("KyberCompress256coeffs", sharing_type.Arithmetic, sharing_type.Boolean, "BronchainC22", KyberCompress256coeffsS2S1_BronchainC22_cost, KyberCompress256coeffsS2S1_BronchainC22_memory, "https://eprint.iacr.org/2022/158.pdf", security_lvl = security_level.PINI)

def KyberCompress256coeffsS2S1_CoronGMZ23_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Compress algorithm 
    of [GMZ22]

    Gives an estimate of the number of CPU cycles of the Compress algorithm 
    of [GMZ22], according to the parameters given at the input of the function. 

    Paper: https://eprint.iacr.org/2021/1615.pdf
    Link: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    """
    KyberCompress_cost = gadgets_dict.get_cost(mask_params, "KyberCompress", sharing_type.Arithmetic, sharing_type.Boolean, name = "CoronGMZ23")
    cost = 256*KyberCompress_cost
    return cost

def KyberCompress256coeffsS2S1_CoronGMZ23_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Compress algorithm 
    of [GMZ22]

    Gives an estimate of the maximum of memory used by the Compress
    algorithm of [GMZ22], according to the parameters given at the input of the 
    function.

    Paper: https://eprint.iacr.org/2021/1615.pdf
    Link: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    """
    memory = gadgets_dict.get_memory(mask_params, "KyberCompress", sharing_type.Arithmetic, sharing_type.Boolean, name = "CoronGMZ23")
    return memory

gadgets_dict.add_new_gadget("KyberCompress256coeffs", sharing_type.Arithmetic, sharing_type.Boolean, "CoronGMZ23", KyberCompress256coeffsS2S1_CoronGMZ23_cost, KyberCompress256coeffsS2S1_CoronGMZ23_memory, "https://eprint.iacr.org/2021/1615.pdf", security_lvl = security_level.NI)

def KyberCompress256coeffsS2S1_CoronGMZ22_13_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the 1-bit Compress algorithm 
    of [GMZ22] (Algorithm 13)

    Gives an estimate of the number of CPU cycles of the 1-bit Compress 
    algorithm of [GMZ22], according to the parameters given at the input of the 
    function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c

    void kyber_decryption(Masked* x, Masked* b){
        unsigned l = switch_table[MASKING_ORDER-1];
        modulus_switch(x, 3329, l);
        convert_2_l_to_1bit_bool(x, b, l);
    }
    """
    d = additional_KyberCompress_params["d"]
    try:
        assert d == 1
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget KyberCompress256coeffsS2S1_CoronGMZ22_13 only works for 1-bit compression, i.e. with the additional parameter d = 1, not {d=}.")
        return None
    KyberCompress_cost = gadgets_dict.get_cost(mask_params, "KyberCompress", sharing_type.Arithmetic, sharing_type.Boolean, name = "CoronGMZ22_13")
    cost = 256*KyberCompress_cost
    return cost

def KyberCompress256coeffsS2S1_CoronGMZ22_13_memory(mask_params):
    """
    Estimation of the maximum of memory used in the 1-bit Compress algorithm of 
    [GMZ22] (Algorithm 13)

    Gives an estimate of the maximum of memory used by the 1-bit Compress 
    algorithm of [GMZ22], according to the parameters given at the input of the 
    function.

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    """
    d = additional_KyberCompress_params["d"]
    try:
        assert d == 1
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget KyberCompress256coeffsS2S1_CoronGMZ22_13 only works for 1-bit compression, i.e. with the additional parameter d = 1, not {d=}.")
        return None
    memory = gadgets_dict.get_memory(mask_params, "KyberCompress", sharing_type.Arithmetic, sharing_type.Boolean, name = "CoronGMZ22_13")
    return memory

gadgets_dict.add_new_gadget("KyberCompress256coeffs", sharing_type.Arithmetic, sharing_type.Boolean, "CoronGMZ22_13", KyberCompress256coeffsS2S1_CoronGMZ22_13_cost, KyberCompress256coeffsS2S1_CoronGMZ22_13_memory, "https://eprint.iacr.org/2021/1314.pdf", security_lvl = security_level.SNI)









