# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import log, ceil

from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

additional_KyberCompress_params = dict()
additional_KyberCompress_params["p"] = 3329
additional_KyberCompress_params["c"] = 1

def KyberCompress32coeffsS2S1_BronchainC22_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the bitsliced Compress 
    algorithm of [BC22]

    Gives an estimate of the number of CPU cycles of the bitsliced Compress 
    algorithm of [BC22], according to the parameters given at the input of the 
    function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void seccompress(size_t nshares, uint32_t q, uint32_t c, uint32_t *out,
                               size_t out_msk_stride, size_t out_data_stride,
                               const int16_t *in, size_t in_msk_stride,
                               size_t in_data_stride)
    """
    
    n = mask_params.n
    p = additional_KyberCompress_params["p"]
    c = additional_KyberCompress_params["c"]
    alpha = ceil(log(p*n, 2))
    mov_op = atomic_operations.get("mov_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("leftshift_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    udiv_op = atomic_operations.get("udiv_op").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    # Computation of alpha
    cost = mult_op
    cost += alpha*(rightshift_op+add_op)
    # Computation of y: mod 2**(alpha+c)
    for i in range(32):
        cost += add_op+modulo_Q_op
        cost += leftshift_op+3*add_op+rightshift_op+udiv_op
        cost += add_op+leftshift_op+sub_op
        cost += and_op+leftshift_op+add_op+sub_op
        for j in range(n):
            cost += add_op+modulo_Q_op
            cost += leftshift_op+3*add_op+rightshift_op+udiv_op
            cost += and_op+leftshift_op+add_op+sub_op
    mask_params.set_bitsize_shares(alpha+c)
    cost += gadgets_dict.get_cost(mask_params,"Dense2bitsliced32bits", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    cost += gadgets_dict.get_cost(mask_params,"BsSecA2BPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    cost += (c*n)*mov_op
    mask_params.set_bitsize_shares(c)
    return cost

def KyberCompress32coeffsS2S1_BronchainC22_memory(mask_params):
    """
    Estimation of the maximum of memory used in the bitsliced Compress 
    algorithm of [BC22]

    Gives an estimate of the maximum of memory used by the bitsliced Compress 
    algorithm of [BC22], according to the parameters given at the input of the 
    function.

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void seccompress(size_t nshares, uint32_t q, uint32_t c, uint32_t *out,
                               size_t out_msk_stride, size_t out_data_stride,
                               const int16_t *in, size_t in_msk_stride,
                               size_t in_data_stride)
    """
    n = mask_params.n
    p = additional_KyberCompress_params["p"]
    c = additional_KyberCompress_params["c"]
    alpha = ceil(log(p*n, 2))
    bitsize_shares = mask_params.bitsize_shares
    # ell
    memory = ceil(log(alpha, 2))
    # in_expanded
    memory += 32*n*(alpha+c)
    # bs_expanded
    memory += (alpha+c)*n*32
    # tmp32, tmp64
    memory += 2*(alpha+c+1)
    mask_params.set_bitsize_shares(alpha+c)
    memory_Dense2bitsliced32bits = gadgets_dict.get_memory(mask_params,"Dense2bitsliced32bits", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    memory_BsSecA2BPowerOf2 = gadgets_dict.get_memory(mask_params, "BsSecA2BPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    memory += max(memory_Dense2bitsliced32bits, memory_BsSecA2BPowerOf2)
    mask_params.set_bitsize_shares(c)
    return memory

gadgets_dict.add_new_gadget("KyberCompress32coeffs", sharing_type.Arithmetic, sharing_type.Boolean, "BronchainC22", KyberCompress32coeffsS2S1_BronchainC22_cost, KyberCompress32coeffsS2S1_BronchainC22_memory, "https://eprint.iacr.org/2022/158.pdf", security_lvl = security_level.PINI)
