# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import log, ceil

from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

delta = 0.02
additional_KyberCompress_params = dict()
additional_KyberCompress_params["q"] = 3329
additional_KyberCompress_params["d"] = 1

def KyberCompressS2S1_CoronGMZ23_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Compress algorithm of 
    [CGMZ23] (Algorithm 2) 

    Gives an estimate of the number of CPU cycles of the Compress algorithm of 
    [CGMZ23], according to the parameters given at the input of the function. 

    Paper: https://eprint.iacr.org/2021/1615.pdf
    Link: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c
    Function: void high_order_compress(Masked* x, Masked* y, unsigned q, unsigned k, unsigned ell)
    Remark: As usual, the values ((1<<(k+ell))-1) and ((1<<k)-1) are considered constants,
            their computation and memory footprint are not considered.
    """
    
    n = mask_params.n
    q = additional_KyberCompress_params["q"]
    d = additional_KyberCompress_params["d"]
    alpha = ceil(log(q*n, 2))
    mov_op = atomic_operations.get("mov_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("leftshift_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    udiv_op = atomic_operations.get("udiv_op").get("cost")

    # compute z.shares[0]
    cost = mov_op
    cost += leftshift_op+2*add_op+udiv_op+mult_op
    cost += add_op+sub_op+leftshift_op+and_op

    # Loop
    cost += n*mov_op
    cost += n*(2*leftshift_op+2*add_op+udiv_op+and_op)

    mask_params.set_bitsize_shares(d+alpha)
    SecA2BModPowerOf2_cost = gadgets_dict.get_cost(mask_params,"SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)

    # in case using BeirendonckDV21_SingleLookUp, it must be initialized
    try:
        assert(gadgets_dict._gadgets_set["SECA2BMODPOWEROF2S2S1"].default_gadget.name != "BeirendonckDV21_SingleLookUp".upper())
    except:
        # Right now we don't know how to make compatible the CoronGMZ23 shares's bit-size inflation 
        # with BeirendonckDV21_SingleLookUp's forced output bit size of 16
        return None

    cost += SecA2BModPowerOf2_cost

    # Loop
    cost += n*(rightshift_op+and_op)
    
    # output Boolean shares
    mask_params.set_bitsize_shares(d)

    return cost

def KyberCompressS2S1_CoronGMZ23_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Compress algorithm of 
    [CGMZ23] (Algorithm 2)

    Gives an estimate of the maximum of memory used by the Compress algorithm of 
    [CGMZ23], according to the parameters given at the input of the function.

    Paper: https://eprint.iacr.org/2021/1615.pdf
    Link: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c
    Function: void high_order_compress(Masked* x, Masked* y, unsigned q, unsigned k, unsigned ell)
    """
    n = mask_params.n
    q = additional_KyberCompress_params["q"]
    d = additional_KyberCompress_params["d"]
    alpha = ceil(log(q*n, 2))
    bitsize_shares = mask_params.bitsize_shares
    # Masked: z, t
    memory = 2*n*(d+alpha)
    # temp
    memory += bitsize_shares
    
    mask_params.set_bitsize_shares(d+alpha)
    SecA2BModPowerOf2_memory = gadgets_dict.get_memory(mask_params, "SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
        
    try:
        assert(gadgets_dict._gadgets_set["SECA2BMODPOWEROF2S2S1"].default_gadget.name != "BeirendonckDV21_SingleLookUp".upper())
    except:
        # Right now we don't know how to make compatible the CoronGMZ23 shares's bit-size inflation 
        # with BeirendonckDV21_SingleLookUp's forced output bit size of 16
        return None

    memory += SecA2BModPowerOf2_memory
    
    # output Boolean shares
    mask_params.set_bitsize_shares(d)
    return memory

gadgets_dict.add_new_gadget("KyberCompress", sharing_type.Arithmetic, sharing_type.Boolean, "CoronGMZ23", KyberCompressS2S1_CoronGMZ23_cost, KyberCompressS2S1_CoronGMZ23_memory, "https://eprint.iacr.org/2021/1615.pdf", security_lvl = security_level.NI)

def KyberCompressS2S1_CoronGMZ22_13_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the 1-bit Compress algorithm 
    of [GMZ22] (Algorithm 13)

    Gives an estimate of the number of CPU cycles of the 1-bit Compress 
    algorithm of [GMZ22], according to the parameters given at the input of the 
    function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function:
    void kyber_decryption(Masked* x, Masked* b){
        unsigned l = switch_table[MASKING_ORDER-1];
        modulus_switch(x, 3329, l);
        convert_2_l_to_1bit_bool(x, b, l);
    }
    """
    
    n = mask_params.n
    d = additional_KyberCompress_params["d"]
    try:
        assert d == 1
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget KyberCompressS2S1_CoronGMZ22_13 only works for 1-bit compression, i.e. with the additional parameter d = 1, not {d=}.")
        return None
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    load_op = atomic_operations.get("load_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    udiv_op = atomic_operations.get("udiv_op").get("cost")
    # Get the value l for the conversion A_{2**l} to boolean of 1 bit
    # l may be greater than 12 because of the way the conversion is performed
    l = ceil(log((n/delta), 2))-1
    cost  = 0
    # Modulus switch between Z_p to Z_{2**l}
    for i in range(n):
        cost += add_op+leftshift_op
        cost += add_op+udiv_op+mult_op
        cost += leftshift_op+sub_op+and_op
    # convert_2_l_to_1bit_bool
    bitsize_shares = mask_params.bitsize_shares
    mask_params.set_bitsize_shares(l)
    cost += gadgets_dict.get_cost(mask_params, "SecA2B1bit", sharing_type.Arithmetic, sharing_type.Boolean, name = "CoronGMZ22_12")
    return cost

def KyberCompressS2S1_CoronGMZ22_13_memory(mask_params):
    """
    Estimation of the maximum of memory used in the 1-bit Compress algorithm of 
    [GMZ22] (Algorithm 13)

    Gives an estimate of the maximum of memory used by the 1-bitCompress
    algorithm of [GMZ22], according to the parameters given at the input of the 
    function.

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function:
    void kyber_decryption(Masked* x, Masked* b){
        unsigned l = switch_table[MASKING_ORDER-1];
        modulus_switch(x, 3329, l);
        convert_2_l_to_1bit_bool(x, b, l);
    }
    """
    
    n = mask_params.n
    d = additional_KyberCompress_params["d"]
    try:
        assert d == 1
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget KyberCompressS2S1_CoronGMZ22_13 only works for 1-bit compression, i.e. with the additional parameter d = 1, not {d=}.")
        return None
    bitsize_shares = mask_params.bitsize_shares
    # Get the value l for the conversion A_{2**l} to boolean of 1 bit
    # l may be greater than 12 because of the way the conversion is performed
    l = ceil(log((n/delta), 2))-1
    memory = ceil(log(l,2))
    memory_modulus_switch = memory
    memory_modulus_switch += bitsize_shares+l+1
    memory_conversionA2B1bit = memory
    mask_params.set_bitsize_shares(l)
    memory_conversionA2B1bit += gadgets_dict.get_memory(mask_params, "SecA2B1bit", sharing_type.Arithmetic, sharing_type.Boolean, name = "CoronGMZ22_12")
    memory = max(memory_modulus_switch, memory_conversionA2B1bit)
    return memory

gadgets_dict.add_new_gadget("KyberCompress", sharing_type.Arithmetic, sharing_type.Boolean, "CoronGMZ22_13", KyberCompressS2S1_CoronGMZ22_13_cost, KyberCompressS2S1_CoronGMZ22_13_memory, "https://eprint.iacr.org/2021/1314.pdf", security_lvl = security_level.SNI)

