# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def MaskExpansionS2S2_SchneiderPOG19_7_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the MaskExpansion algorithm 
    of [SPOG19] (Algorithm 7)

    Gives an estimate of the number of CPU cycles of the MaskExpansion algorithm 
    of [SPOG19] (Algorithm 7), according to the parameters given at the input of 
    the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910.pdf

    Reference code:
    (From: https://github.com/uclcrypto/pqm4_masked/files/8048895/implems.zip)

    void sec_b2a_qbit_n(uint32_t *c,const uint32_t *a,const uint32_t xn,const uint32_t n){
	    uint32_t b[n];
	    uint32_t r;
	    size_t i,j;

	    b[n-1] = sampleq();
	    b[0] = (a[0] - b[n-1] + Q)%Q;

	    for(j=1;j<n-1;j++){
	        r = sampleq();
	        b[j] = (a[j] - r + Q)%Q;
	        b[n-1] = (b[n-1] + r)%Q;
	    }

	    for(i=0;i<n;i++){
	        c[i] = b[i] + 2*Q;
	        c[i] -= (2*b[i]*xn);
	        c[i] = c[i]%Q;
	    }
	    c[0] = (c[0] + xn)%Q;
	}
    """
    mask_params.new_number_of_shares(mask_params.n+1)
    n = mask_params.n
    rand_generation_prime_field = atomic_operations.get("rand_generation_prime_field").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    add_op = (atomic_operations.get("add_op")).get("cost")
    sub_op = (atomic_operations.get("sub_op")).get("cost")
    mult_op = (atomic_operations.get("mult_op")).get("cost")
    cost = rand_generation_prime_field
    cost += sub_op+add_op+modulo_Q_op
    for j in range(1,n-1):
    	cost += rand_generation_prime_field
    	cost += sub_op+add_op+modulo_Q_op
    	cost += add_op+modulo_Q_op
    for i in range(n):
    	cost += add_op+mult_op
    	cost += sub_op+2*mult_op
    	cost += modulo_Q_op
    cost += add_op+modulo_Q_op
    return cost

def MaskExpansionS2S2_SchneiderPOG19_7_memory(mask_params):
    """
    Estimation of the maximum of memory used in the MaskExpansion algorithm of 
    [SPOG19] (Algorithm 7)

    Gives an estimate of the maximum of memory used by the MaskExpansion 
    algorithm of [SPOG19] (Algorithm 7), according to the parameters given at 
    the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910

    Reference code:
    (From: https://github.com/uclcrypto/pqm4_masked/files/8048895/implems.zip)

    void sec_b2a_qbit_n(uint32_t *c,const uint32_t *a,const uint32_t xn,const uint32_t n){
	    uint32_t b[n];
	    uint32_t r;
	    size_t i,j;

	    b[n-1] = sampleq();
	    b[0] = (a[0] - b[n-1] + Q)%Q;

	    for(j=1;j<n-1;j++){
	        r = sampleq();
	        b[j] = (a[j] - r + Q)%Q;
	        b[n-1] = (b[n-1] + r)%Q;
	    }

	    for(i=0;i<n;i++){
	        c[i] = b[i] + 2*Q;
	        c[i] -= (2*b[i]*xn);
	        c[i] = c[i]%Q;
	    }
	    c[0] = (c[0] + xn)%Q;
	}
    """
    mask_params.new_number_of_shares(mask_params.n+1)
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    # uint32_t: b[n], r
    memory = n*bitsize_shares+bitsize_shares
    return memory

gadgets_dict.add_new_gadget("MaskExpansion", sharing_type.Arithmetic, sharing_type.Arithmetic, "SchneiderPOG19_7", MaskExpansionS2S2_SchneiderPOG19_7_cost, MaskExpansionS2S2_SchneiderPOG19_7_memory, "https://eprint.iacr.org/2019/910", security_lvl = security_level.NI)

