# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def RefreshS1S1_BartheBDFGSZ16_4b_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Refresh algorithm (4b) 
    of [BBDFGSZ16]

    Gives an estimate of the number of CPU cycles of the Refresh algorithm (4b) 
    of [BBDFGSZ16], according to the parameters given at the input of the 
    function.

    Paper: Strong Non-Interference and Type-Directed Higher-Order Masking
    Link: https://eprint.iacr.org/2015/506.pdf

    Reference code:

    void RefreshXOR(Masked x, uint32_t order, uint32_t w){
        # Computation of the mask for the modulo 2**w operation
        uint32_t r;
        uint32_t mask;
        mask = (1<<w) -1; // assumed to be constant, not counted in cost or mem
        # Randomization of each part of the masked variable done in two nested for loops
        for(int i = 0; i<(order-1); i++){
            for(int j = i+1; j<order; j++){
                r=rand();
                r = r & mask;
                x[i]=x[i]^r;
                x[j]=x[j]^r;
            }
        }
    }
    """
    
    # Retrieval of the number of shares of masked values in gadget input
    n = mask_params.n
    
    # Retrieving the cost of all types of operations performed in our code
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")

    # Randomization of each part of the masked variable done in two nested for loops
    cost = ((n*(n-1))/2)*(rand_generation+and_op+(2*xor_op))
    return cost

def RefreshS1S1_BartheBDFGSZ16_4b_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Refresh algorithm (4b) of 
    [BBDFGSZ16]

    Gives an estimate of the maximum of memory used by the Refresh algorithm 
    (4b) of [BBDFGSZ16], according to the parameters given at the input of the 
    function.

    Paper: Strong Non-Interference and Type-Directed Higher-Order Masking
    Link: https://eprint.iacr.org/2015/506.pdf

    Reference code:

    void RefreshXOR(Masked x, uint32_t order, uint32_t w){
        uint32_t r;
        uint32_t mask;
        mask = (1<<w) -1; // assumed to be constant, not counted in cost or mem
        for(int i = 0; i<(order-1); i++){
            for(int j = i+1; j<order; j++){
                r=rand();
                r = r & mask;
                x[i]=x[i]^r;
                x[j]=x[j]^r;
            }
        }
    }
    """
    # Gadget memory cost computation
    # uint32_t: r
    memory = mask_params.bitsize_shares
    return memory

gadgets_dict.add_new_gadget("Refresh", sharing_type.Boolean, sharing_type.Boolean, "BartheBDFGSZ16_4b", RefreshS1S1_BartheBDFGSZ16_4b_cost, RefreshS1S1_BartheBDFGSZ16_4b_memory, "https://eprint.iacr.org/2015/506.pdf", security_lvl = security_level.SNI)

def RefreshS1S1_RivainP10_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Refresh algorithm of 
    [RP10]

    Gives an estimate of the number of CPU cycles of the Refresh algorithm of 
    [RP10], according to the parameters given at the input of the function. 

    Paper: Provably Secure Higher-Order Masking of AES
    Link: https://eprint.iacr.org/2010/441.pdf 

    Reference code: https://eprint.iacr.org/2010/441.pdf 
                    -> Algorithm 4: RefreshMasks
    """
    
    n = mask_params.n
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")

    # Randomization of each part of the masked variable done in two nested for loops
    cost = (n-1)*(rand_generation+and_op+(2*xor_op))
    return cost

def RefreshS1S1_RivainP10_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Refresh algorithm of 
    [RP10]

    Gives an estimate of the maximum of memory used by the Refresh algorithm of 
    [RP10], according to the parameters given at the input of the function. 

    Paper: Provably Secure Higher-Order Masking of AES
    Link: https://eprint.iacr.org/2010/441.pdf 

    Reference code: https://eprint.iacr.org/2010/441.pdf 
                    -> Algorithm 4: RefreshMasks
    """
    # uint32_t: r
    memory = mask_params.bitsize_shares
    return memory

gadgets_dict.add_new_gadget("Refresh", sharing_type.Boolean, sharing_type.Boolean, "RivainP10", RefreshS1S1_RivainP10_cost, RefreshS1S1_RivainP10_memory, "https://eprint.iacr.org/2010/441.pdf", security_lvl = security_level.NI)


