# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def RefreshS2S2_BartheBDFGSZ16_4b_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Refresh algorithm (4b) 
    of [BBDFGSZ16]

    Gives an estimate of the number of CPU cycles of the Refresh algorithm (4b) 
    of [BBDFGSZ16], according to the parameters given at the input of the 
    function.

    Paper: Strong Non-Interference and Type-Directed Higher-Order Masking
    Link: https://eprint.iacr.org/2015/506.pdf

    Reference code:

    void refresh_add(uint32_t *a){
        size_t i,j;
        uint32_t rnd;
        for(i=0;i<D-1;i++){
            for(j=i+1;j<D;j++){
                rnd = sampleq();
                a[i] = (a[i] + rnd);
                a[j] = (a[j] + Q - rnd);
            }
        }
        for(i=0;i<D;i++){
            a[i] = a[i]%KYBER_Q;
        }
    }
    """
    
    # Retrieval of the number of shares of masked values in gadget input
    n = mask_params.n
    
    # Retrieving the cost of all types of operations performed in our code
    rand_generation_prime_field = atomic_operations.get("rand_generation_prime_field").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")

    # Randomization of each part of the masked variable done in two nested for loops
    cost = ((n*(n-1))/2)*(rand_generation_prime_field+(2*add_op)+sub_op)
    cost += n*modulo_Q_op
    return cost

def RefreshS2S2_BartheBDFGSZ16_4b_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Refresh algorithm (4b) of 
    [BBDFGSZ16]

    Gives an estimate of the maximum of memory used by the Refresh algorithm 
    (4b) of [BBDFGSZ16], according to the parameters given at the input of the 
    function.

    Paper: Strong Non-Interference and Type-Directed Higher-Order Masking
    Link: https://eprint.iacr.org/2015/506.pdf

    Reference code:

    void refresh_add(uint32_t *a){
        size_t i,j;
        uint32_t rnd;
        for(i=0;i<D-1;i++){
            for(j=i+1;j<D;j++){
                rnd = sampleq();
                a[i] = (a[i] + rnd);
                a[j] = (a[j] + Q - rnd);
            }
        }
        for(i=0;i<D;i++){
            a[i] = a[i]%KYBER_Q;
        }
    }
    """
    # Gadget memory cost computation
    # uint32_t: r
    memory = mask_params.bitsize_shares
    return memory

gadgets_dict.add_new_gadget("Refresh", sharing_type.Arithmetic, sharing_type.Arithmetic, "BartheBDFGSZ16_4b", RefreshS2S2_BartheBDFGSZ16_4b_cost, RefreshS2S2_BartheBDFGSZ16_4b_memory, "https://eprint.iacr.org/2015/506.pdf", security_lvl = security_level.SNI)
