# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def Sec1bitB2AModpS1S2_SchneiderPOG19_6_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Sec1bitB2AModp algorithm 
    of [SPOG19] (Algorithm 6)

    Gives an estimate of the number of CPU cycles of the Sec1bitB2AModp 
    algorithm of [SPOG19] (Algorithm 6), according to the parameters given at 
    the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910.pdf

    Reference code:
    (From: https://github.com/uclcrypto/pqm4_masked/files/8048895/implems.zip)

    void b2a_qbit(uint32_t *a,uint32_t *x){
	    a[0] = x[0];
	    for(int i=1;i<D;i++){
	        id = sec_b2a_qbit_n(a,a,x[i],i+1);
	    }
	}
    """
    n = mask_params.n
    mov_op = atomic_operations.get("mov_op").get("cost")
    cost = mov_op
    mask_params.new_number_of_shares(1)
    for i in range(1,n):
    	cost += gadgets_dict.get_cost(mask_params,"MaskExpansion", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "SchneiderPOG19_7")
    return cost

def Sec1bitB2AModpS1S2_SchneiderPOG19_6_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Sec1bitB2AModp algorithm of 
    [SPOG19] (Algorithm 6)

    Gives an estimate of the maximum of memory used by the Sec1bitB2AModp 
    algorithm of [SPOG19] (Algorithm 6), according to the parameters given at 
    the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910

    Reference code:
    (From: https://github.com/uclcrypto/pqm4_masked/files/8048895/implems.zip)

    void b2a_qbit(uint32_t *a,uint32_t *x){
	    a[0] = x[0];
	    for(int i=1;i<D;i++){
	        sec_b2a_qbit_n(a,a,x[i],i+1);
	    }
	}
    """
    mask_params.new_number_of_shares(mask_params.n-1)
    bitsize_shares = mask_params.bitsize_shares
    memory = gadgets_dict.get_memory(mask_params,"MaskExpansion", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "SchneiderPOG19_7")
    return memory

gadgets_dict.add_new_gadget("Sec1bitB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, "SchneiderPOG19_6", Sec1bitB2AModpS1S2_SchneiderPOG19_6_cost, Sec1bitB2AModpS1S2_SchneiderPOG19_6_memory, "https://eprint.iacr.org/2019/910", security_lvl = security_level.NI)

def Sec1bitB2AModpS1S2_SchneiderPOG19_5_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Sec1bitB2AModp algorithm 
    of [SPOG19] (Algorithm 5)

    Gives an estimate of the number of CPU cycles of the Sec1bitB2AModp 
    algorithm of [SPOG19] (Algorithm 5), according to the parameters given at 
    the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910.pdf

    Reference code:
    (From: https://github.com/uclcrypto/pqm4_masked/files/8048895/implems.zip)

    void sec_b2a_qbit(uint32_t *a, uint32_t *x){
	    b2a_qbit(a,x);
	    refresh_add(a);
	}
    """
    cost = gadgets_dict.get_cost(mask_params,"Sec1bitB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, name = "SchneiderPOG19_6")
    cost += gadgets_dict.get_cost(mask_params,"Refresh", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "BartheBDFGSZ16_4b")
    return cost

def Sec1bitB2AModpS1S2_SchneiderPOG19_5_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Sec1bitB2AModp algorithm of 
    [SPOG19] (Algorithm 5)

    Gives an estimate of the maximum of memory used by the Sec1bitB2AModp 
    algorithm of [SPOG19] (Algorithm 5), according to the parameters given at 
    the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910

    Reference code:
    (From: https://github.com/uclcrypto/pqm4_masked/files/8048895/implems.zip)

    void sec_b2a_qbit(uint32_t *a, uint32_t *x){
	    b2a_qbit(a,x);
	    refresh_add(a);
	}
    """
    memory_1bitB2A = gadgets_dict.get_memory(mask_params,"Sec1bitB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, name = "SchneiderPOG19_6")
    memory_Refresh = gadgets_dict.get_memory(mask_params,"Refresh", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "BartheBDFGSZ16_4b")
    memory = max(memory_1bitB2A, memory_Refresh)
    return memory

gadgets_dict.add_new_gadget("Sec1bitB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, "SchneiderPOG19_5", Sec1bitB2AModpS1S2_SchneiderPOG19_5_cost, Sec1bitB2AModpS1S2_SchneiderPOG19_5_memory, "https://eprint.iacr.org/2019/910", security_lvl = security_level.SNI)

