# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def SecA2B1bitS2S1_CoronGMZ22_12_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2B1bit algorithm 
    of [GMZ22] (Algorithm 12)

    Gives an estimate of the number of CPU cycles of the SecA2B1bit algorithm of 
    [GMZ22], according to the parameters given at the input of the function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function: void convert_2_l_to_1bit_bool(Masked* x, Masked* b, unsigned l)
    """
    
    n = mask_params.n
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    load_op = atomic_operations.get("load_op").get("cost")
    store_op = atomic_operations.get("store_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    
    # Generation table
    cost = store_op
    cost += (n-1)*store_op
    for i in range(n-1):
        # Cyclic rotation of the registers to the right
        cost += n*(load_op+sub_op+leftshift_op+rightshift_op+add_op+store_op)
        # Random generation
        cost += (n-1)*rand_generation
        # Randomization of all the shares
        cost += (n-1)*(2*(load_op+xor_op+store_op))
    cost += n*(load_op+rightshift_op+and_op)
    # Final Refresh
    cost += (n-1)*(2*(xor_op+store_op)+rand_generation)
    return cost

def SecA2B1bitS2S1_CoronGMZ22_12_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2B1bit algorithm 
    of [GMZ22] (Algorithm 12)

    Gives an estimate of the maximum of memory used by the SecA2B1bit algorithm 
    of [GMZ22], according to the parameters given at the input of the function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function: void convert_2_l_to_1bit_bool(Masked* x, Masked* b, unsigned l)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    # Computation of the memory taken by the variables
    # uint32_t: r
    memory = bitsize_shares
    # Table
    memory += n*(2**bitsize_shares)
    return memory

gadgets_dict.add_new_gadget("SecA2B1bit", sharing_type.Arithmetic, sharing_type.Boolean, "CoronGMZ22_12", SecA2B1bitS2S1_CoronGMZ22_12_cost, SecA2B1bitS2S1_CoronGMZ22_12_memory, "https://eprint.iacr.org/2021/1314.pdf", security_lvl = security_level.SNI)
