# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import floor, ceil, log
from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

additional_SecA2BModPowerOf2_params = dict()
additional_SecA2BModPowerOf2_params["a2b"] = 2
additional_SecA2BModPowerOf2_params["k"] = 8

def SecA2BModPowerOf2InitS2S1_BeirendonckDV21_SingleLookUp_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2BModPowerOf2 
    initialization algorithm of [BDV21]

    Gives an estimate of the number of CPU cycles of the SecA2BModPowerOf2 
    initialization algorithm of [BDV21], according to the parameters given at 
    the input of the function. 

    Paper: Analysis and Comparison of Table-based Arithmetic to Boolean Masking
    Link: https://tches.iacr.org/index.php/TCHES/article/view/8975/8553

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/common/a2b_singlelookup.c
    Function: void A2B_init(void)
    """
    
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget SecA2BModPowerOf2InitS2S1_BeirendonckDV21_SingleLookUp only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    a2bn = additional_SecA2BModPowerOf2_params.get("a2b")
    k = additional_SecA2BModPowerOf2_params.get("k")
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    store_op = atomic_operations.get("store_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    or_op = atomic_operations.get("or_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    pow2 = leftshift_op
    mod2 = pow2+sub_op
    # Generation of rho and the buffer to construct each r_i and 
    # rrr = (r_n || ... || r_i || ... || r_0) 
    cost = 2*rand_generation+and_op
    # Construction of each r_i
    cost += a2bn*(rightshift_op+mult_op+and_op+mod2)
    # Construction of (r_n || ... || r_i || ... || r_0)
    cost += a2bn*(leftshift_op+mult_op+or_op)
    # Computation of tables T_i
    # Indexes index(i, rho, A)
    cost += a2bn*(2**k)*(add_op+xor_op+leftshift_op+or_op+store_op)
    # Indexes index(i, (rho ^ 1) , A)
    cost += a2bn*(2**k)*(2*add_op+2*xor_op+leftshift_op+or_op+store_op)
    return cost

def SecA2BModPowerOf2InitS2S1_BeirendonckDV21_SingleLookUp_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2BModPowerOf2 
    initialization algorithm of [BDV21]

    Gives an estimate of the maximum of memory used by the SecA2BModPowerOf2 
    initialization algorithm of [BDV21], according to the parameters given at 
    the input of the function.

    Paper: Analysis and Comparison of Table-based Arithmetic to Boolean Masking
    Link: https://tches.iacr.org/index.php/TCHES/article/view/8975/8553

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/common/a2b_singlelookup.c
    Function: void A2B_init(void)
    """
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget SecA2BModPowerOf2InitS2S1_BeirendonckDV21_SingleLookUp only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    a2bn = additional_SecA2BModPowerOf2_params.get("a2b")
    k = additional_SecA2BModPowerOf2_params.get("k")
    # Computation of the memory taken by the variables
    # 1 bit: rho
    memory = 1
    # k bits: A, r
    memory += k+a2bn*k
    # 2*k bits: buff
    memory += 2*k
    # A2B*k bits: rrr
    memory += a2bn*k
    # T
    memory += a2bn*(2**(k+1))*a2bn*k
    return memory

gadgets_dict.add_new_gadget("SecA2BModPowerOf2Init", sharing_type.Arithmetic, sharing_type.Boolean, "BeirendonckDV21_SingleLookUp", SecA2BModPowerOf2InitS2S1_BeirendonckDV21_SingleLookUp_cost, SecA2BModPowerOf2InitS2S1_BeirendonckDV21_SingleLookUp_memory, "https://tches.iacr.org/index.php/TCHES/article/view/8975/8553", security_lvl = None)

def SecA2BModPowerOf2S2S1_BeirendonckDV21_SingleLookUp_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2BModPowerOf2 
    algorithm of [BDV21]

    Gives an estimate of the number of CPU cycles of the SecA2BModPowerOf2 
    algorithm of [BDV21], according to the parameters given at the input of the 
    function. 

    Paper: Analysis and Comparison of Table-based Arithmetic to Boolean Masking
    Link: https://tches.iacr.org/index.php/TCHES/article/view/8975/8553

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/common/a2b_singlelookup.c
    Function: uint32_t A2B_convert(uint32_t A, uint32_t R)
              without the Init function
    """
    
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget SecA2BModPowerOf2S2S1_BeirendonckDV21_SingleLookUp only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    a2bn = additional_SecA2BModPowerOf2_params.get("a2b")
    k = additional_SecA2BModPowerOf2_params.get("k")
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    load_op = atomic_operations.get("load_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    or_op = atomic_operations.get("or_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    pow2 = leftshift_op
    mod2 = pow2+sub_op
    # A2B_convert_sw
    cost = sub_op+and_op
    # for loop
    # Computation of A_l
    cost += a2bn*(3*and_op+add_op+sub_op+mult_op+mod2)
    # Computation of beta and B_i
    cost += a2bn*(load_op+and_op+rightshift_op+xor_op)
    # Computation of B
    cost += a2bn*(leftshift_op+mult_op+or_op)
    # Shift of A and R for the loop
    cost += a2bn*(2*rightshift_op)
    # Result returned
    cost += xor_op
    return cost

def SecA2BModPowerOf2S2S1_BeirendonckDV21_SingleLookUp_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2BModPowerOf2 
    algorithm of [BDV21]

    Gives an estimate of the maximum of memory used by the SecA2BModPowerOf2 
    algorithm of [BDV21], according to the parameters given at the input of the 
    function.

    Paper: Analysis and Comparison of Table-based Arithmetic to Boolean Masking
    Link: https://tches.iacr.org/index.php/TCHES/article/view/8975/8553

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/common/a2b_singlelookup.c
    Function: uint32_t A2B_convert(uint32_t A, uint32_t R)
              without the Init function
    """
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget SecA2BModPowerOf2S2S1_BeirendonckDV21_SingleLookUp only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    a2bn = additional_SecA2BModPowerOf2_params.get("a2b")
    k = additional_SecA2BModPowerOf2_params.get("k")
    # Computation of the memory taken by the variables
    # A_l, R_l, betaBi, beta, Bi, B
    memory = 2*k+(k+1)+(k+1)+a2bn*k
    # T
    memory += a2bn*(2**(k+1))*a2bn*k
    return memory

gadgets_dict.add_new_gadget("SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, "BeirendonckDV21_SingleLookUp", SecA2BModPowerOf2S2S1_BeirendonckDV21_SingleLookUp_cost, SecA2BModPowerOf2S2S1_BeirendonckDV21_SingleLookUp_memory, "https://tches.iacr.org/index.php/TCHES/article/view/8975/8553", security_lvl = None)

def SecA2BModPowerOf2S2S1_CoronGV14_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2BModPowerOf2 
    algorithm of [CGV14] (Algorithm 4)

    Gives an estimate of the number of CPU cycles of the SecA2BModPowerOf2 
    algorithm of [CGV14], according to the parameters given at the input of the 
    function. 

    Paper: Secure Conversion Between Boolean and Arithmetic Masking of Any Order
    Link: http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf

    Reference code: None
    """
    
    n = mask_params.n

    # force number of shares to be a power of 2
    if (n & (n-1)) != 0:
        n_powerof2 = 2
        while n_powerof2<n:
            n_powerof2 *= 2
        n = n_powerof2
        mask_params.new_number_of_shares(n)

    if(n == 1):
        return 0
    else:
        rightshift_op = atomic_operations.get("rightshift_op").get("cost")
        SecAdd_cost = gadgets_dict.get_cost(mask_params, "SecAddModPowerOf2", sharing_type.Boolean, sharing_type.Boolean, name = None)
        cost = rightshift_op
        mask_params_copy = deepcopy(mask_params)
        half_order = n//2
        mask_params_copy.new_number_of_shares(half_order)
        cost += 2*SecA2BModPowerOf2S2S1_CoronGV14_cost(mask_params_copy)
        cost += 2*gadgets_dict.get_cost(mask_params_copy, "Expand", sharing_type.Boolean, sharing_type.Boolean, name = "CoronGV14")
        cost += SecAdd_cost
    return cost

def SecA2BModPowerOf2S2S1_CoronGV14_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2BModPowerOf2 
    algorithm of [CGV14] (Algorithm 4)

    Gives an estimate of the maximum of memory used by the SecA2BModPowerOf2 
    algorithm of [CGV14], according to the parameters given at the input of the 
    function. 

    Paper: Secure Conversion Between Boolean and Arithmetic Masking of Any Order
    Link: http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf

    Reference code: None
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares

    if (n & (n-1)) != 0:
        n_powerof2 = 2
        while n_powerof2<n:
            n_powerof2 *= 2
        n = n_powerof2
        mask_params.new_number_of_shares(n)
    
    if(n == 1):
        memory = 0
    else:
        # Computation of the memory taken by the variables
        half_order = n//2
        # Masked: x, xp,y, yp
        memory = 2*half_order*bitsize_shares+2*n*bitsize_shares
        # uint32_t: half_order
        memory += ceil(log(half_order, 2))

        # recurcive calls
        mask_params_copy = deepcopy(mask_params)
        mask_params_copy.n = half_order
        memory_1 = SecA2BModPowerOf2S2S1_CoronGV14_memory(mask_params_copy)
        mask_params_copy2 = deepcopy(mask_params)
        mask_params_copy2.n = n-half_order
        memory_2 = SecA2BModPowerOf2S2S1_CoronGV14_memory(mask_params_copy2)
        
        # SecAdd
        SecAdd_memory = gadgets_dict.get_memory(mask_params, "SecAddModPowerOf2", sharing_type.Boolean, sharing_type.Boolean, name = None)
        
        memory += max(memory_1, memory_2, SecAdd_memory)

    return memory

gadgets_dict.add_new_gadget("SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, "CoronGV14", SecA2BModPowerOf2S2S1_CoronGV14_cost, SecA2BModPowerOf2S2S1_CoronGV14_memory, "http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf", security_lvl = security_level.NI)

def SecA2BModPowerOf2S2S1_Goubin01_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2BModPowerOf2 
    algorithm of [Goubin01] (Algorithm 2)

    Gives an estimate of the number of CPU cycles of the SecA2BModPowerOf2 
    algorithm of [Goubin01], according to the parameters given at the input of 
    the function. 

    Paper: A Sound Method for Switching between Boolean and Arithmetic Masking
    Link: http://www.goubin.fr/papers/arith-final.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c
    Function: uint32_t GoubinAB(uint32_t A,uint32_t r,int k)
    """
    
    n = mask_params.n
    k = mask_params.bitsize_shares
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget SecA2BModPowerOf2S2S1_Goubin01 only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")

    cost = rand_generation
    cost += leftshift_op
    cost += xor_op
    cost += and_op
    cost += 4*xor_op+2*and_op
    cost += (k-1)*(2*and_op+2*xor_op+leftshift_op)
    cost += xor_op
    return cost

def SecA2BModPowerOf2S2S1_Goubin01_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2BModPowerOf2 
    algorithm of [Goubin01]

    Gives an estimate of the maximum of memory used by the SecA2BModPowerOf2 
    algorithm of [Goubin01], according to the parameters given at the input of 
    the function.

    Paper: A Sound Method for Switching between Boolean and Arithmetic Masking
    Link: http://www.goubin.fr/papers/arith-final.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c
    Function: uint32_t GoubinAB(uint32_t A,uint32_t r,int k)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget SecA2BModPowerOf2S2S1_Goubin01 only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    # uint32_t: G, T, x, O
    memory = 4*bitsize_shares
    return memory

gadgets_dict.add_new_gadget("SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, "Goubin01", SecA2BModPowerOf2S2S1_Goubin01_cost, SecA2BModPowerOf2S2S1_Goubin01_memory, "http://www.goubin.fr/papers/arith-final.pdf", security_lvl = None)








