# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import ceil, log, floor

from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def SecA2BModpS2S1_BartheBEFGRT18_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2BModp algorithm 
    of [BBEFGRT18] (Algorithm 12)

    Gives an estimate of the number of CPU cycles of the SecA2BModp algorithm of 
    [BBEFGRT18], according to the parameters given at the input of the function. 

    Paper: Masking the GLP Lattice-Based Signature Scheme at Any Order
    Link: https://eprint.iacr.org/2018/381.pdf
    Algorithm 12

    Reference code:

    void SecArithBoolModp_cubic(Masked a, Masked b, uint32_t p, uint32_t order, uint32_t w){
        Masked y_shared = malloc(order*sizeof(uint32_t));
        Masked tmp_shared = malloc(order*sizeof(uint32_t));
        for(int i=0; i<order; i++){
            b[i] = 0;
        }
        for(int j=0; j<order; j++){
            y_shared[0]=a[j];
            for(int i=1; i<order; i++){
                y_shared[i]=0;
            }
            RefreshXOR(y_shared, order, w);
            SecAddModp(b, y_shared, tmp_shared, p, order, w);
            for(int i=0; i<order; i++){
                b[i]=tmp_shared[i];
            }
        }
        free(y_shared);
        free(tmp_shared);
    }

    Complexity: O(n**3)
    """
    
    n = mask_params.n
    Refresh_cost = gadgets_dict.get_cost(mask_params, "Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAddModp_cost = gadgets_dict.get_cost(mask_params, "SecAddModp", sharing_type.Boolean, sharing_type.Boolean, name = None)
    cost = 0
    for j in range(n):
        # y = Refresh(y)
        cost += Refresh_cost
        # tmp = b + y mod p
        cost += SecAddModp_cost
    return cost

def SecA2BModpS2S1_BartheBEFGRT18_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2BModp algorithm 
    of [BBEFGRT18] (Algorithm 12)

    Gives an estimate of the maximum of memory used by the SecA2BModp algorithm 
    of [BBEFGRT18], according to the parameters given at the input of the 
    function.

    Paper: Masking the GLP Lattice-Based Signature Scheme at Any Order
    Link: https://eprint.iacr.org/2018/381.pdf

    Reference code:

    void SecArithBoolModp_cubic(Masked a, Masked b, uint32_t p, uint32_t order, uint32_t w){
        Masked y_shared = malloc(order*sizeof(uint32_t));
        Masked tmp_shared = malloc(order*sizeof(uint32_t));
        for(int i=0; i<order; i++){
            b[i] = 0;
        }
        for(int j=0; j<order; j++){
            y_shared[0]=a[j];
            for(int i=1; i<order; i++){
                y_shared[i]=0;
            }
            RefreshXOR(y_shared, order, w);
            SecAddModp(b, y_shared, tmp_shared, p, order, w);
            for(int i=0; i<order; i++){
                b[i]=tmp_shared[i];
            }
        }
        free(y_shared);
        free(tmp_shared);
    }

    Complexity: O(n**3)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    SecAddModp_memory = gadgets_dict.get_memory(mask_params, "SecAddModp", sharing_type.Boolean, sharing_type.Boolean, name = None)
    Refresh_memory    = gadgets_dict.get_memory(mask_params, "Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    # Computation of the memory taken by the variables
    # Masked: y_shared, tmp_shared
    memory = 2*n*bitsize_shares
    memory += max(SecAddModp_memory, Refresh_memory)
    return memory

gadgets_dict.add_new_gadget("SecA2BModp", sharing_type.Arithmetic, sharing_type.Boolean, "BartheBEFGRT18", SecA2BModpS2S1_BartheBEFGRT18_cost, SecA2BModpS2S1_BartheBEFGRT18_memory, "https://eprint.iacr.org/2018/381.pdf", security_lvl = security_level.SNI)

def SecA2BModpS2S1_SchneiderPOG19_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecA2BModp algorithm 
    of [SPOG19] (Algorithm 3)

    Gives an estimate of the number of CPU cycles of the SecA2BModp algorithm of 
    [SPOG19], according to the parameters given at the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910.pdf

    Reference code:

    void SecArithBoolModp_quadratic(Masked a, Masked b, uint32_t p, uint32_t order, uint32_t w){
        if(order == 1){
            b[0]=a[0];
        }
        else{
            uint32_t half_order = order >> 1;
            Masked y_shared = malloc(order*sizeof(uint32_t));
            Masked z_shared = malloc(order*sizeof(uint32_t));
            Masked tmp_shared = malloc((order-half_order)*sizeof(uint32_t));
            SecArithBoolModp_quadratic(a, y_shared, p, half_order, w);
            for(int i=half_order; i<order; i++){
                y_shared[i]=0;
            }
            RefreshXOR(y_shared, order, w);
            for(int i=half_order; i<order; i++){
                tmp_shared[i-half_order]=a[i];
            }
            SecArithBoolModp_quadratic(tmp_shared, z_shared, p, (order-half_order), w);
            for(int i=(order-half_order); i<order; i++){
                z_shared[i]=0;
            }
            RefreshXOR(z_shared, order, w);
            SecAddModp(y_shared, z_shared, b, p, order, w);
            free(y_shared);
            free(z_shared);
            free(tmp_shared);
        }
    }

    Complexity: O(n**2 log_2(k))
    """
    
    n = mask_params.n
    if(n == 1):
        return 0
    else:
        rightshift_op = atomic_operations.get("rightshift_op").get("cost")
        Refresh_cost = gadgets_dict.get_cost(mask_params, "Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
        SecAddModp_cost = gadgets_dict.get_cost(mask_params, "SecAddModp", sharing_type.Boolean, sharing_type.Boolean, name = None)
        cost = rightshift_op
        mask_params_copy = deepcopy(mask_params)
        half_order = floor(n/2)
        mask_params_copy.n = half_order
        cost += SecA2BModpS2S1_SchneiderPOG19_cost(mask_params_copy)
        cost += Refresh_cost
        mask_params_copy2 = deepcopy(mask_params)
        mask_params_copy2.n = n-half_order
        cost += SecA2BModpS2S1_SchneiderPOG19_cost(mask_params_copy2)
        cost += Refresh_cost
        cost += SecAddModp_cost
    return cost

def SecA2BModpS2S1_SchneiderPOG19_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecA2BModp algorithm of 
    [SPOG19]

    Gives an estimate of the maximum of memory used by the SecA2BModp algorithm 
    of [SPOG19], according to the parameters given at the input of the function. 

    Paper: Efficiently Masking Binomial Sampling at Arbitrary Orders for 
    Lattice-Based Crypto
    Link: https://eprint.iacr.org/2019/910

    Reference code:

    void SecArithBoolModp_quadratic(Masked a, Masked b, uint32_t p, uint32_t order, uint32_t w){
        if(order == 1){
            b[0]=a[0];
        }
        else{
            uint32_t half_order = order >> 1;
            Masked y_shared = malloc(order*sizeof(uint32_t));
            Masked z_shared = malloc(order*sizeof(uint32_t));
            Masked tmp_shared = malloc((order-half_order)*sizeof(uint32_t));
            SecArithBoolModp_quadratic(a, y_shared, p, half_order, w);
            for(int i=half_order; i<order; i++){
                y_shared[i]=0;
            }
            RefreshXOR(y_shared, order, w);
            for(int i=half_order; i<order; i++){
                tmp_shared[i-half_order]=a[i];
            }
            SecArithBoolModp_quadratic(tmp_shared, z_shared, p, (order-half_order), w);
            for(int i=(order-half_order); i<order; i++){
                z_shared[i]=0;
            }
            RefreshXOR(z_shared, order, w);
            SecAddModp(y_shared, z_shared, b, p, order, w);
            free(y_shared);
            free(z_shared);
            free(tmp_shared);
        }
    }

    Complexity: O(n**2 log_2(k))
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    if(n == 1):
        memory = 0
    else:
        # Computation of the memory taken by the variables
        half_order = n//2
        # Masked: y_shared, z_shared, tmp_shared
        memory = 2*n*bitsize_shares+(n-half_order)*bitsize_shares
        # uint32_t: half_order
        memory += ceil(log(half_order, 2))

        # recurcive calls
        mask_params_copy = deepcopy(mask_params)
        mask_params_copy.n = half_order
        memory_1 = SecA2BModpS2S1_SchneiderPOG19_memory(mask_params_copy)
        mask_params_copy2 = deepcopy(mask_params)
        mask_params_copy2.n = n-half_order
        memory_2 = SecA2BModpS2S1_SchneiderPOG19_memory(mask_params_copy2)
        
        # SecAdd
        SecAddModp_memory = gadgets_dict.get_memory(mask_params, "SecAddModp", sharing_type.Boolean, sharing_type.Boolean, name = None)
        
        memory += max(memory_1, memory_2, SecAddModp_memory)

    return memory

gadgets_dict.add_new_gadget("SecA2BModp", sharing_type.Arithmetic, sharing_type.Boolean, "SchneiderPOG19", SecA2BModpS2S1_SchneiderPOG19_cost, SecA2BModpS2S1_SchneiderPOG19_memory, "https://eprint.iacr.org/2019/910", security_lvl = security_level.SNI)


