# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import ceil, log
from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level


def SecAddModPowerOf2S1S1_CoronGTV15_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecAdd algorithm of 
    [CGTV15] (Kogge-Stone Adder)

    Gives an estimate of the number of CPU cycles of the SecAdd algorithm of 
    [CGTV15], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Conversion from Arithmetic to Boolean Masking with Logarithmic 
           Complexity
    Link: https://eprint.iacr.org/2014/891.pdf

    Reference code:

    void SecAdd(Masked x, Masked y, Masked result, uint32_t order, uint32_t w){
        uint32_t power, W;
        uint32_t mask;
        Masked p_shared = malloc(order*sizeof(uint32_t));
        Masked g_shared = malloc(order*sizeof(uint32_t));
        Masked a_shared = malloc(order*sizeof(uint32_t));
        Masked aprime_shared = malloc(order*sizeof(uint32_t));
        Masked tmp_shared = malloc(order*sizeof(uint32_t));
        mask = (1<<w) - 1; // assumed to be constant, not counted in cost or mem
        // P_0 = x ^ y
        for(int i=0; i<order; i++){
            p_shared[i] = x[i]^y[i];
        }
        // G_0 = x & y
        SecAnd(x, y, g_shared, order, w);
        W = (uint32_t)(ceil(log2(w-1))-1);
        // Computing P_i and G_i
        for(int j=1; j<=W; j++){
            // power = (uint32_t)(pow(2, j-1));
            power = 1 << (j-1);
            // G_{i-1} << 2^(i-1)
            for(int i=0; i<order; i++){
                a_shared[i] = (g_shared[i]<<power);
            }
            // P_{i-1} & (G_{i-1} << 2^(i-1))
            SecAnd(a_shared, p_shared, tmp_shared, order, w);
            for(int i=0; i<order; i++){
                a_shared[i] = tmp_shared[i];
            }
            // G_i = (P_{i-1} & (G_{i-1} << 2^(i-1))) ^ G_{i-1}
            // P_{i-1} << 2^(i-1)
            for(int i=0; i<order; i++){
                g_shared[i] ^= a_shared[i];
                aprime_shared[i] = (p_shared[i]<<power);
            }
            RefreshXOR(aprime_shared, order, w);
            // P_i = P_{i-1} & (P_{i-1} << 2^(i-1))
            SecAnd(p_shared, aprime_shared, tmp_shared, order, w);
            for(int i=0; i<order; i++){
                p_shared[i] = tmp_shared[i];
            }
        }
        // G_W << 2^W
        power = 1 << W;
        for(int i=0; i<order; i++){
            a_shared[i] = (g_shared[i]<<power);
        }
        // P_W & (G_W << 2^W)
        SecAnd(a_shared, p_shared, tmp_shared, order, w);
        for(int i=0; i<order; i++){
            a_shared[i] = tmp_shared[i];
        }
        // G_{W+1} = (P_W & (G_W << 2^W)) ^ G_W
        for(int i=0; i<order; i++){
            g_shared[i] ^= a_shared[i];
        }
        // x + y mod 2^k = x ^ y ^ (2*G_{W+1})
        for(int i=0; i<order; i++){
            result[i] = x[i]^y[i];
            result[i] ^= (g_shared[i]<<1)&mask;
        }
        free(p_shared);
        free(g_shared);
        free(a_shared);
        free(aprime_shared);
        free(tmp_shared);
    }
    """
    

    n = mask_params.n
    W = ceil(log(mask_params.bitsize_shares-1,2))-1
    rand_generation = (atomic_operations.get("rand_generation")).get("cost")
    load_op = (atomic_operations.get("load_op")).get("cost")
    leftshift_op = (atomic_operations.get("leftshift_op")).get("cost")
    xor_op = (atomic_operations.get("xor_op")).get("cost")
    and_op = (atomic_operations.get("and_op")).get("cost")
    sub_op = (atomic_operations.get("sub_op")).get("cost")
    Refresh_cost = gadgets_dict.get_cost(mask_params,"Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAnd_cost = gadgets_dict.get_cost(mask_params, "SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    
    # P_0 = x ^ y
    cost = n*xor_op
    # G_0 = x & y
    cost += SecAnd_cost
    # Computing P_i and G_i
    for i in range(W):
        # G_{i-1} << 2^(i-1)
        cost += sub_op+leftshift_op
        cost += n*leftshift_op
        # P_{i-1} & (G_{i-1} << 2^(i-1))
        cost += SecAnd_cost
        # G_i = (P_{i-1} & (G_{i-1} << 2^(i-1))) ^ G_{i-1}
        # P_{i-1} << 2^(i-1)
        cost += n*(xor_op+leftshift_op)
        cost += Refresh_cost
        # P_i = P_{i-1} & (P_{i-1} << 2^(i-1))
        cost += SecAnd_cost
    # G_W << 2^W
    cost += leftshift_op
    cost += n*leftshift_op
    # P_W & (G_W << 2^W)
    cost += SecAnd_cost
    # G_{W+1} = (P_W & (G_W << 2^W)) ^ G_W
    cost += n*xor_op
    # x + y mod 2^k = x ^ y ^ (2*G_{W+1})
    cost += n*(2*xor_op+leftshift_op+and_op)
    return cost

def SecAddModPowerOf2S1S1_CoronGTV15_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecAdd algorithm of 
    [CGTV15] (Kogge-Stone Adder)

    Gives an estimate of the maximum of memory used by the SecAdd algorithm of 
    [CGTV15], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Conversion from Arithmetic to Boolean Masking with Logarithmic 
           Complexity
    Link: https://eprint.iacr.org/2014/891.pdf

    Reference code:

    void SecAdd(Masked x, Masked y, Masked result, uint32_t order, uint32_t w){
        uint32_t power, W;
        uint32_t mask;
        Masked p_shared = malloc(order*sizeof(uint32_t));
        Masked g_shared = malloc(order*sizeof(uint32_t));
        Masked a_shared = malloc(order*sizeof(uint32_t));
        Masked aprime_shared = malloc(order*sizeof(uint32_t));
        Masked tmp_shared = malloc(order*sizeof(uint32_t));
        mask = (1<<w) - 1; // assumed to be constant, not counted in cost or mem
        // P_0 = x ^ y
        for(int i=0; i<order; i++){
            p_shared[i] = x[i]^y[i];
        }
        // G_0 = x & y
        SecAnd(x, y, g_shared, order, w);
        W = (uint32_t)(ceil(log2(w-1))-1);
        // Computing P_i and G_i
        for(int j=1; j<=W; j++){
            // power = (uint32_t)(pow(2, j-1));
            power = 1 << (j-1);
            // G_{i-1} << 2^(i-1)
            for(int i=0; i<order; i++){
                a_shared[i] = (g_shared[i]<<power);
            }
            // P_{i-1} & (G_{i-1} << 2^(i-1))
            SecAnd(a_shared, p_shared, tmp_shared, order, w);
            for(int i=0; i<order; i++){
                a_shared[i] = tmp_shared[i];
            }
            // G_i = (P_{i-1} & (G_{i-1} << 2^(i-1))) ^ G_{i-1}
            // P_{i-1} << 2^(i-1)
            for(int i=0; i<order; i++){
                g_shared[i] ^= a_shared[i];
                aprime_shared[i] = (p_shared[i]<<power);
            }
            RefreshXOR(aprime_shared, order, w);
            // P_i = P_{i-1} & (P_{i-1} << 2^(i-1))
            SecAnd(p_shared, aprime_shared, tmp_shared, order, w);
            for(int i=0; i<order; i++){
                p_shared[i] = tmp_shared[i];
            }
        }
        // G_W << 2^W
        power = 1 << W;
        for(int i=0; i<order; i++){
            a_shared[i] = (g_shared[i]<<power);
        }
        // P_W & (G_W << 2^W)
        SecAnd(a_shared, p_shared, tmp_shared, order, w);
        for(int i=0; i<order; i++){
            a_shared[i] = tmp_shared[i];
        }
        // G_{W+1} = (P_W & (G_W << 2^W)) ^ G_W
        for(int i=0; i<order; i++){
            g_shared[i] ^= a_shared[i];
        }
        // x + y mod 2^k = x ^ y ^ (2*G_{W+1})
        for(int i=0; i<order; i++){
            result[i] = x[i]^y[i];
            result[i] ^= (g_shared[i]<<1)&mask;
        }
        free(p_shared);
        free(g_shared);
        free(a_shared);
        free(aprime_shared);
        free(tmp_shared);
    }
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    W = ceil(log(mask_params.bitsize_shares-1,2))-1
    bitsize_W = ceil(log(W, 2))
    Refresh_memory = gadgets_dict.get_memory(mask_params,"Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAnd_memory = gadgets_dict.get_memory(mask_params, "SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    # Computation of the space taken by the variables
    # Masked: p_shared, g_shared, a_shared, aprime_shared, tmp_shared
    memory = 5*n*bitsize_shares
    # uint32_t: power, W
    memory += W+bitsize_W
    # adding sub-operation with maximum memory footprint
    memory += max(Refresh_memory, SecAnd_memory)
    return memory

gadgets_dict.add_new_gadget("SecAddModPowerOf2", sharing_type.Boolean, sharing_type.Boolean, "CoronGTV15", SecAddModPowerOf2S1S1_CoronGTV15_cost, SecAddModPowerOf2S1S1_CoronGTV15_memory, "https://eprint.iacr.org/2014/891.pdf", security_lvl = security_level.NI)

def SecAddModPowerOf2S1S1_CoronGV14_3_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecAdd algorithm of 
    [CGV14]

    Gives an estimate of the number of CPU cycles of the SecAdd algorithm of 
    [CGV14], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Secure Conversion Between Boolean and Arithmetic Masking of Any Order
    Link: http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c
    Function: void SecAdd(uint32_t *x,uint32_t *y,uint32_t *z,int k,int n)
    """
    
    n = mask_params.n
    k = mask_params.bitsize_shares
    leftshift_op = (atomic_operations.get("leftshift_op")).get("cost")
    xor_op = (atomic_operations.get("xor_op")).get("cost")
    and_op = (atomic_operations.get("and_op")).get("cost")
    sub_op = (atomic_operations.get("sub_op")).get("cost")
    mult_op = (atomic_operations.get("mult_op")).get("cost")
    SecAnd_cost = gadgets_dict.get_cost(mask_params, "SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    
    cost = SecAnd_cost
    cost += n*xor_op
    cost += (k-1)*SecAnd_cost
    cost += (k-1)*n*(xor_op+leftshift_op+and_op) # the "2*" is assumed to be a leftshift_op
    cost += n*2*xor_op

    return cost

def SecAddModPowerOf2S1S1_CoronGV14_3_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecAdd algorithm of 
    [CGV14]

    Gives an estimate of the maximum of memory used by the SecAdd algorithm of 
    [CGV14], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Secure Conversion Between Boolean and Arithmetic Masking of Any Order
    Link: http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/Masked_KEMs/Kyber/polycomp.c
    Function: void SecAdd(uint32_t *x,uint32_t *y,uint32_t *z,int k,int n)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares

    SecAnd_memory = gadgets_dict.get_memory(mask_params, "SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    # uint32_t: u, w, a, ua
    memory = 4*n*bitsize_shares
    memory += SecAnd_memory
    return memory

gadgets_dict.add_new_gadget("SecAddModPowerOf2", sharing_type.Boolean, sharing_type.Boolean, "CoronGV14_3", SecAddModPowerOf2S1S1_CoronGV14_3_cost, SecAddModPowerOf2S1S1_CoronGV14_3_memory, "http://www.crypto-uni.lu/jscoron/publications/secconvorder.pdf", security_lvl =  None)


