# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import ceil, log
from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def SecAddModpS1S1_BartheBEFGRT18_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecAddModp algorithm 
    of [BBEFGRT18]

    Gives an estimate of the number of CPU cycles of the SecAddModp algorithm of 
    [BBEFGRT18], according to the parameters given at the input of the function. 
    The number of bits used as input to the function is taken into account in 
    the calculations.

    Paper: Masking the GLP Lattice-Based Signature Scheme at Any Order
    Link: https://eprint.iacr.org/2018/381.pdf

    Reference code:

    void SecAddModq(uint32_t *x,
            uint32_t *y,
            uint32_t *z,
            uint32_t d,
            uint32_t q,
            uint32_t k){
        uint32_t p[d];
        uint32_t sp[d];
        uint32_t s[d];
        uint32_t b[d];
        uint32_t *c;
        uint32_t cp[d];
        uint32_t zp[d];

        uint32_t mask = (1<<k)-1; // assumed to be constant, not counted in cost or mem
        for(uint32_t i=1;i<d;i++){
            p[i] = 0;
        }
        p[0] = (1<<k) - q; // assumed to be constant, not counted in cost or mem

        SecAdd(x,y,s,d,k);
        SecAdd(s,p,sp,d,k);
        
        for(uint32_t i=0;i<d;i++){
            b[i] = sp[i] >> (k-1);
        }
        
        c = b;
        if(k>16){RefreshXOR(c,d);}else{RefreshXOR16(c,d);};
        for(uint32_t i=0;i<d;i++){
           cp[i] = ((1<<k) - (c[i]&0x1))&mask; 
        }

        if(k>16){SecAnd(s,cp,z,d);}else{SecAnd16(s,cp,z,d);}

        if(k>16){RefreshXOR(c,d);}else{RefreshXOR16(c,d);};
        c[0] ^= 1;
        for(uint32_t i=0;i<d;i++){
           cp[i] = ((1<<k) - (c[i]&0x1))&mask; 
        }

        if(k>16){SecAnd(sp,cp,zp,d);}else{SecAnd16(sp,cp,zp,d);}
        for(uint32_t i=0; i<d;i++){
            z[i] ^= zp[i];
        }
    }
    """
    global scheme_related_parameters, atomic_operations
    n = mask_params.n
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    Refresh_cost = gadgets_dict.get_cost(mask_params, "Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAnd_cost = gadgets_dict.get_cost(mask_params, "SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAdd_cost = gadgets_dict.get_cost(mask_params, "SecAddModPowerOf2", sharing_type.Boolean, sharing_type.Boolean, name = None)
    # s = x+y mod 2**w
    cost = SecAdd_cost
    # sp = s+p mod 2**w
    cost += SecAdd_cost
    # b = MSB(sp)
    cost += n*(rightshift_op+sub_op)
    # c = RefreshXOR(b)
    cost += Refresh_cost
    # Extension of bit b over the w bits of cp
    cost += n*(leftshift_op+sub_op+2*and_op)
    # z = s & cp
    cost += SecAnd_cost
    # c = RefreshXOR(c)
    cost += Refresh_cost
    # c = not(c)
    cost += xor_op
    # Extension of bit of not(b) over the w bits of cp
    cost += n*(leftshift_op+sub_op+2*and_op)
    # zp = sp & cp
    cost += SecAnd_cost
    # z[i] ^= zp[i]
    cost += n*xor_op
    return cost

def SecAddModpS1S1_BartheBEFGRT18_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecAddModp algorithm of 
    [BBEFGRT18]

    Gives an estimate of the maximum of memory used by the SecAddModp algorithm 
    of [BBEFGRT18], according to the parameters given at the input of the 
    function. The number of bits used as input to the function is taken into 
    account in the calculations.

    Paper: Masking the GLP Lattice-Based Signature Scheme at Any Order
    Link: https://eprint.iacr.org/2018/381.pdf

    Reference code:

    void SecAddModq(uint32_t *x,
            uint32_t *y,
            uint32_t *z,
            uint32_t d,
            uint32_t q,
            uint32_t k){
        uint32_t p[d];
        uint32_t sp[d];
        uint32_t s[d];
        uint32_t b[d];
        uint32_t *c;
        uint32_t cp[d];
        uint32_t zp[d];

        uint32_t mask = (1<<k)-1; // assumed to be constant, not counted in cost or mem
        for(uint32_t i=1;i<d;i++){
            p[i] = 0;
        }
        p[0] = (1<<k) - q; // assumed to be constant, not counted in cost or mem

        SecAdd(x,y,s,d,k);
        SecAdd(s,p,sp,d,k);
        
        for(uint32_t i=0;i<d;i++){
            b[i] = sp[i] >> (k-1);
        }
        
        c = b;
        if(k>16){RefreshXOR(c,d);}else{RefreshXOR16(c,d);};
        for(uint32_t i=0;i<d;i++){
           cp[i] = ((1<<k) - (c[i]&0x1))&mask; 
        }

        if(k>16){SecAnd(s,cp,z,d);}else{SecAnd16(s,cp,z,d);}

        if(k>16){RefreshXOR(c,d);}else{RefreshXOR16(c,d);};
        c[0] ^= 1;
        for(uint32_t i=0;i<d;i++){
           cp[i] = ((1<<k) - (c[i]&0x1))&mask; 
        }

        if(k>16){SecAnd(sp,cp,zp,d);}else{SecAnd16(sp,cp,zp,d);}
        for(uint32_t i=0; i<d;i++){
            z[i] ^= zp[i];
        }
    }
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    
    Refresh_memory = gadgets_dict.get_memory(mask_params, "Refresh", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAnd_memory = gadgets_dict.get_memory(mask_params, "SecAnd", sharing_type.Boolean, sharing_type.Boolean, name = None)
    SecAdd_memory = gadgets_dict.get_memory(mask_params, "SecAddModPowerOf2", sharing_type.Boolean, sharing_type.Boolean, name = None)
    # Computation of the memory taken by the variables
    # Masked values of w bits: p, sp, s, b, cp, zp
    memory = 6*n*bitsize_shares
    memory += max(Refresh_memory, SecAnd_memory, SecAdd_memory)
    return memory

gadgets_dict.add_new_gadget("SecAddModp", sharing_type.Boolean, sharing_type.Boolean, "BartheBEFGRT18", SecAddModpS1S1_BartheBEFGRT18_cost, SecAddModpS1S1_BartheBEFGRT18_memory, "https://eprint.iacr.org/2018/381.pdf", security_lvl = security_level.NI)

