# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def SecAndS1S1_IshaiSW03_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecAnd algorithm of 
    [ISW03]

    Gives an estimate of the number of CPU cycles of the SecAnd algorithm of 
    [ISW03], according to the parameters given at the input of the function. The 
    number of bits used as input to the function is taken into account in the 
    calculations.

    Paper: Private Circuits: Securing Hardware against Probing Attacks
    Link: https://people.eecs.berkeley.edu/~daw/papers/privcirc-crypto03.pdf

    Reference code:

    void SecAnd(Masked a, Masked b, Masked result, uint32_t order, uint32_t w){
        uint32_t rij, rji, tmp;
        uint32_t mask = (1<<w) - 1; // this is constant, not counted in cost or mem
        for(int i = 0; i<order; i++){
            result[i]=a[i]&b[i];
        }
        for(int i=0; i<(order-1); i++){
            for(int j=i+1; j<order; j++){
                rij=rand();
                rij=rij & mask;
                rji=a[i]&b[j];
                rji^=rij;
                tmp=a[j]&b[i];
                rji^=tmp;
                result[i]^=rij;
                result[j]^=rji;
            }
        }
    }
    """
    
    n = mask_params.n
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    mask_op = atomic_operations.get("mask_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    # First loop computing a&b
    cost = n*and_op
    # Computations done in the two nested loops
    cost += ((n*(n-1))/2)*(rand_generation+(3*and_op)+(4*xor_op))
    return cost

def SecAndS1S1_IshaiSW03_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecAnd algorithm of 
    [ISW03]

    Gives an estimate of the maximum of memory used by the SecAnd algorithm of 
    [ISW03], according to the parameters given at the input of the function. The 
    number of bits used as input to the function is taken into account in the 
    calculations.

    Paper: Private Circuits: Securing Hardware against Probing Attacks
    Link: https://people.eecs.berkeley.edu/~daw/papers/privcirc-crypto03.pdf

    Reference code:

    void SecAnd(Masked a, Masked b, Masked result, uint32_t order, uint32_t w){
        uint32_t rij, rji, tmp;
        uint32_t mask = (1<<w) - 1; // this is constant, not counted in cost or mem
        for(int i = 0; i<order; i++){
            result[i]=a[i]&b[i];
        }
        for(int i=0; i<(order-1); i++){
            for(int j=i+1; j<order; j++){
                rij=rand();
                rij=rij & mask;
                rji=a[i]&b[j];
                rji^=rij;
                tmp=a[j]&b[i];
                rji^=tmp;
                result[i]^=rij;
                result[j]^=rji;
            }
        }
    }
    """
    # uint32_t: rij, rji, tmp 
    # they are all of minimum size mask_params.bitsize_shares
    memory = 3*mask_params.bitsize_shares
    return memory

gadgets_dict.add_new_gadget("SecAnd", sharing_type.Boolean, sharing_type.Boolean, "IshaiSW03", SecAndS1S1_IshaiSW03_cost, SecAndS1S1_IshaiSW03_memory, "https://people.eecs.berkeley.edu/~daw/papers/privcirc-crypto03.pdf", security_lvl =  security_level.SNI)


