# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def SecB2AModpS1S2_CoronGMZ22_3_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecB2AModp algorithm 
    of [CGMZ22] (Algorithm 3)

    Gives an estimate of the number of CPU cycles of the SecB2AModp 
    algorithm of [CGMZ22] (Algorithm 3), according to the parameters given at 
    the input of the function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function: void convert_B2A(Masked* x, Masked* y, unsigned k, unsigned q)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    load_op = atomic_operations.get("load_op").get("cost")
    store_op = atomic_operations.get("store_op").get("cost")
    Refresh_cost = gadgets_dict.get_cost(mask_params,"Refresh", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "BartheBDFGSZ16_4b")
    tab_size = 2**bitsize_shares
    cost = tab_size*(modulo_Q_op+store_op)
    cost += tab_size*(n-1)*store_op
    cost += (n-1)*tab_size*n*(load_op+store_op)
    cost += (n-1)*tab_size*(load_op+Refresh_cost+store_op)
    cost += (n-1)*tab_size*n*(load_op+store_op)
    cost += n*load_op
    cost += Refresh_cost
    return cost

def SecB2AModpS1S2_CoronGMZ22_3_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecB2AModp algorithm of 
    [CGMZ22] (Algorithm 3)

    Gives an estimate of the maximum of memory used by the SecB2AModp 
    algorithm of [CGMZ22] (Algorithm 3), according to the parameters given at 
    the input of the function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function: void convert_B2A(Masked* x, Masked* y, unsigned k, unsigned q)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    tab_size = 2**bitsize_shares
    # Masked T[(1<<k)], T_p[(1<<k)]
    memory = 2*tab_size*n*bitsize_shares
    memory += gadgets_dict.get_memory(mask_params,"Refresh", sharing_type.Arithmetic, sharing_type.Arithmetic, name = "BartheBDFGSZ16_4b")
    return memory

gadgets_dict.add_new_gadget("SecB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, "CoronGMZ22_3", SecB2AModpS1S2_CoronGMZ22_3_cost, SecB2AModpS1S2_CoronGMZ22_3_memory, "https://eprint.iacr.org/2021/1314.pdf", security_lvl = security_level.SNI)

def SecB2AModpS1S2_CoronGMZ22_5_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecB2AModp algorithm 
    of [CGMZ22] (Algorithm 5)

    Gives an estimate of the number of CPU cycles of the SecB2AModp 
    algorithm of [CGMZ22] (Algorithm 5), according to the parameters given at 
    the input of the function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function: void optimized_convert_B2A(Masked* x, Masked* y, unsigned k, unsigned q)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    mask_params.set_bitsize_shares(1)
    SecB2AModp_cost = gadgets_dict.get_cost(mask_params,"SecB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, name = "CoronGMZ22_3")
    mask_params.set_bitsize_shares(bitsize_shares)
    tab_size = 2**bitsize_shares
    cost = bitsize_shares*n*(rightshift_op+and_op)
    cost += bitsize_shares*SecB2AModp_cost
    cost += bitsize_shares*n*(leftshift_op+add_op+modulo_Q_op)
    return cost

def SecB2AModpS1S2_CoronGMZ22_5_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecB2AModp algorithm of 
    [CGMZ22] (Algorithm 5)

    Gives an estimate of the maximum of memory used by the SecB2AModp 
    algorithm of [CGMZ22] (Algorithm 5), according to the parameters given at 
    the input of the function. 

    Paper: High-order Table-based Conversion Algorithms and Masking 
    Lattice-based Encryption
    Link: https://eprint.iacr.org/2021/1314.pdf

    Reference code: https://github.com/fragerar/HOTableConv/blob/main/TableConv/new_gadgets.c
    Function: void optimized_convert_B2A(Masked* x, Masked* y, unsigned k, unsigned q)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    tab_size = 2**bitsize_shares
    # Masked z: n shares of 1 bit
    memory = n
    # Masked t: n arithmetic shares mod q
    memory += n*bitsize_shares
    mask_params.set_bitsize_shares(1)
    memory += gadgets_dict.get_memory(mask_params,"SecB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, name = "CoronGMZ22_3")
    mask_params.set_bitsize_shares(bitsize_shares)
    return memory

gadgets_dict.add_new_gadget("SecB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, "CoronGMZ22_5", SecB2AModpS1S2_CoronGMZ22_5_cost, SecB2AModpS1S2_CoronGMZ22_5_memory, "https://eprint.iacr.org/2021/1314.pdf", security_lvl = security_level.SNI)

def SecB2AModpS1S2_BronchainC22_11_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecB2AModp algorithm 
    of [CGMZ22] (Algorithm 11)

    Gives an estimate of the number of CPU cycles of the SecB2AModp 
    algorithm of [CGMZ22] (Algorithm 11), according to the parameters given at 
    the input of the function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void secb2a_modp(size_t nshares,
                               //   size_t kbits, // MUST BE EQUAL TO COEF_NBITS
                               uint32_t p, uint32_t *in, size_t in_msk_stride,
                               size_t in_data_stride)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    rand_generation_prime_field = atomic_operations.get("rand_generation_prime_field").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    cost = (n-1)*64*(rand_generation_prime_field+sub_op)
    mask_params.set_bitsize_shares(12)
    bitsize_shares = mask_params.bitsize_shares
    mask_params.new_number_of_shares(n-1)
    cost += gadgets_dict.get_cost(mask_params,"Dense2bitsliced32bits", sharing_type.Arithmetic, sharing_type.Arithmetic, name = None)
    mask_params.new_number_of_shares(n)
    cost += 2*gadgets_dict.get_cost(mask_params,"BsSecA2BPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    cost += 2*gadgets_dict.get_cost(mask_params,"BsSecA2BPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    return cost

def SecB2AModpS1S2_BronchainC22_11_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecB2AModp algorithm of 
    [CGMZ22] (Algorithm 11)

    Gives an estimate of the maximum of memory used by the SecB2AModp 
    algorithm of [CGMZ22] (Algorithm 11), according to the parameters given at 
    the input of the function. 

    Paper: Bitslicing Arithmetic/Boolean Masking Conversions for Fun and Profit
    Link: https://eprint.iacr.org/2022/158.pdf

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/gadgets.c
    Function: void secb2a_modp(size_t nshares,
                               //   size_t kbits, // MUST BE EQUAL TO COEF_NBITS
                               uint32_t p, uint32_t *in, size_t in_msk_stride,
                               size_t in_data_stride)
    """
    n = mask_params.n
    bitsize_shares = mask_params.bitsize_shares
    tab_size = 2**bitsize_shares
    # Masked z: n shares of 1 bit
    memory = n
    # Masked t: n arithmetic shares mod q
    memory += n*bitsize_shares
    mask_params.set_bitsize_shares(1)
    memory += gadgets_dict.get_memory(mask_params,"SecB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, name = "CoronGMZ22_3")
    mask_params.set_bitsize_shares(bitsize_shares)
    return memory

gadgets_dict.add_new_gadget("SecB2AModp", sharing_type.Boolean, sharing_type.Arithmetic, "BronchainC22_11", SecB2AModpS1S2_BronchainC22_11_cost, SecB2AModpS1S2_BronchainC22_11_memory, "https://eprint.iacr.org/2021/1314.pdf", security_lvl = security_level.PINI)



