# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def SecMultModpS2S2_IshaiSW03_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the SecAnd algorithm of 
    [ISW03]

    Gives an estimate of the number of CPU cycles of the SecAnd algorithm of 
    [ISW03], according to the parameters given at the input of the function. The 
    number of bits used as input to the function is taken into account in the 
    calculations.

    Paper: Private Circuits: Securing Hardware against Probing Attacks
    Link: https://people.eecs.berkeley.edu/~daw/papers/privcirc-crypto03.pdf

    Reference code:

    void SecAnd(Masked a, Masked b, Masked result, uint32_t order, uint32_t Q){
        uint32_t rij, rji, tmp;
        for(int i = 0; i<order; i++){
            result[i]=(a[i]*b[i]) % Q;
        }
        for(int i=0; i<(order-1); i++){
            for(int j=i+1; j<order; j++){
                rij=rand()%Q;
                rji=a[i]*b[j];
                rji+=rij;
                tmp=a[j]*b[i];
                rji+=tmp;
                rij = result[i] + Q - rij
                rji += result[j]
                result[i] = rij %Q;
                result[j] = rji %Q;
            }
        }
    }
    """
    
    n = mask_params.n
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    # First loop computing a*b
    cost = n*mult_op * n*modulo_Q_op
    # Computations done in the two nested loops
    cost += ((n*(n-1))/2) * (rand_generation+(2*mult_op)+(4*add_op)+sub_op+(3*modulo_Q_op))
    return cost

def SecMultModpS2S2_IshaiSW03_memory(mask_params):
    """
    Estimation of the maximum of memory used in the SecAnd algorithm of 
    [ISW03]

    Gives an estimate of the maximum of memory used by the SecAnd algorithm of 
    [ISW03], according to the parameters given at the input of the function. The 
    number of bits used as input to the function is taken into account in the 
    calculations.

    Paper: Private Circuits: Securing Hardware against Probing Attacks
    Link: https://people.eecs.berkeley.edu/~daw/papers/privcirc-crypto03.pdf

    Reference code:

    void SecAnd(Masked a, Masked b, Masked result, uint32_t order, uint32_t Q){
        uint32_t rij, rji, tmp;
        for(int i = 0; i<order; i++){
            result[i]=(a[i]*b[i]) % Q;
        }
        for(int i=0; i<(order-1); i++){
            for(int j=i+1; j<order; j++){
                rij=rand()%Q;
                rji=a[i]*b[j];
                rji+=rij;
                tmp=a[j]*b[i];
                rji+=tmp;
                rij = result[i] + Q - rij
                rji += result[j]
                result[i] = rij %Q;
                result[j] = rji %Q;
            }
        }
    }
    """
    # tmp is of minimum size 2*mask_params.bitsize_shares
    # rij is of minimum size mask_params.bitsize_shares+1
    # rji is of minimum size 2*mask_params.bitsize_shares+2
    memory = 5*mask_params.bitsize_shares + 3
    return memory

gadgets_dict.add_new_gadget("SecMultModp", sharing_type.Arithmetic, sharing_type.Arithmetic, "IshaiSW03", SecMultModpS2S2_IshaiSW03_cost, SecMultModpS2S2_IshaiSW03_memory, "https://people.eecs.berkeley.edu/~daw/papers/privcirc-crypto03.pdf", security_lvl =  security_level.SNI)

