# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def TransformPower2S2S2_OderSPG18_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the TransformPower2 
    algorithm of [OSPG18]

    Gives an estimate of the number of CPU cycles of the TransformPower2 
    algorithm of [OSPG18], according to the parameters given at the input of the 
    function. 

    Paper: Practical CCA2-Secure and Masked Ring-LWE Implementation
    Link: https://tches.iacr.org/index.php/TCHES/article/view/836/788

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/crypto_kem/kyber768/m4/masked-poly.c
    Function: void masked_poly_tomsg(masked_u8_msgbytes* msg, masked_poly* a)
    Remark: in masked_poly_tomsg, all Kyber's compression is done, 
            we hence do not consider lines 105-106 and lines 131 to 136 to match 
            with TransformPower2 algorithm of [OSPG18]
    """
    
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget TransformPower2S2S2_OderSPG18 only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    rand_generation = atomic_operations.get("rand_generation").get("cost")
    modulo_Q_op = atomic_operations.get("modulo_Q_op").get("cost")
    load_op = atomic_operations.get("load_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    add_op = atomic_operations.get("add_op").get("cost")
    sub_op = atomic_operations.get("sub_op").get("cost")
    mult_op = atomic_operations.get("mult_op").get("cost")
    SecA2BModPowerOf2_cost = gadgets_dict.get_cost(mask_params, "SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    
    # Generation of y_1 and random
    cost = rand_generation+2*and_op+rightshift_op
    # Computation of y_2
    cost += sub_op+add_op
    # Computation of z0 and z
    cost += sub_op+SecA2BModPowerOf2_cost
    # Generation of k_11 and k_21
    cost += rand_generation+2*and_op+rightshift_op
    # Computation of k_12 and k_22
    cost += 2*rightshift_op+xor_op+2*sub_op
    # Computation of c[0]
    cost += 5*add_op+2*sub_op+2*rightshift_op+xor_op+14*mult_op
    # Computation of c[1]
    cost += sub_op
    return cost

def TransformPower2S2S2_OderSPG18_memory(mask_params):
    """
    Estimation of the maximum of memory used in the TransformPower2 
    algorithm of [OSPG18]

    Gives an estimate of the maximum of memory used by the TransformPower2 
    algorithm of [OSPG18], according to the parameters given at the input of the 
    function.

    Paper: Practical CCA2-Secure and Masked Ring-LWE Implementation
    Link: https://tches.iacr.org/index.php/TCHES/article/view/836/788

    Reference code: https://github.com/masked-kyber-m4/mkm4/blob/main/crypto_kem/kyber768/m4/masked-poly.c
    Function: void masked_poly_tomsg(masked_u8_msgbytes* msg, masked_poly* a)
    Remark: in masked_poly_tomsg, all Kyber's compression is done, 
            we hence do not consider lines 105-106 and lines 131 to 136 to match 
            with TransformPower2 algorithm of [OSPG18]
    """
    n = mask_params.n
    try:
        assert n == 2
    except AssertionError:
        if gadgets_dict.INTERNAL_VERBOSE:
            print(f"The gadget TransformPower2S2S2_OderSPG18 only works at first order, i.e. with a number of shares n equal to 2, not {n=}.")
        return None
    # Computation of the memory taken by the variables
    # uint32_t: tmp
    memory = 32
    # uint16_t: k11, k12, k21, k22, y1, y2, z0, z[0], z[1], random
    memory += 10*16
    memory += gadgets_dict.get_memory(mask_params, "SecA2BModPowerOf2", sharing_type.Arithmetic, sharing_type.Boolean, name = None)
    return memory

gadgets_dict.add_new_gadget("TransformPower2", sharing_type.Arithmetic, sharing_type.Arithmetic, "OderSPG18", TransformPower2S2S2_OderSPG18_cost, TransformPower2S2S2_OderSPG18_memory, "https://tches.iacr.org/index.php/TCHES/article/view/836/788", security_lvl =  None)


