# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from ..atomic_operations import *
from ..parameters import *
from .. import gadgets_dict
from ..gadgets_dict import sharing_type, security_level

def Transpose32S2S2_Warren2002_cost(mask_params):
    """
    Estimation of the number of CPU cycles made by the Transpose32 algorithm
    of [Warren2002] 

    Gives an estimate of the number of CPU cycles of the Transpose32 algorithm 
    of [Warren2002], according to the parameters given at the input of the 
    function.

    Book: "Hacker's Delight" by Henry S. Warren, Jr.
    Link: http://www.icodeguru.com/Embedded/Hacker's-Delight/

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/masked_representations.c
    Function: void transpose32(uint32_t a[32])
    """
    
    rightshift_op = atomic_operations.get("rightshift_op").get("cost")
    leftshift_op = atomic_operations.get("leftshift_op").get("cost")
    xor_op = atomic_operations.get("xor_op").get("cost")
    and_op = atomic_operations.get("and_op").get("cost")
    # Loop computing the transposed 32 bits matrix
    cost = 0
    m = 0x0000FFFF
    j = 16
    while j != 0:
        k = 0
        while k<32:
            cost += xor_op+rightshift_op+and_op
            cost += xor_op
            cost += xor_op+leftshift_op
            k = (k + j + 1) & ~j
        j = j >> 1
        m = m ^ (m << j)
    return cost

def Transpose32S2S2_Warren2002_memory(mask_params):
    """
    Estimation of the maximum of memory used in the Transpose32 algorithm of 
    [Warren2002]

    Gives an estimate of the maximum of memory used by the Transpose32 
    algorithm of [Warren2002], according to the parameters given at the input of 
    the function. 

    Book: "Hacker's Delight" by Henry S. Warren, Jr.
    Link: http://www.icodeguru.com/Embedded/Hacker's-Delight/

    Reference code: https://github.com/uclcrypto/pqm4_masked/blob/master/crypto_kem/kyber768/m4fspeed_masked/masked_representations.c
    Function: void transpose32(uint32_t a[32])
    """
    # uint32_t: m, t
    # even if m, t are defined over 32 bits, it can be check that they necessitate only 16 bits.
    memory = 2*16
    return memory

gadgets_dict.add_new_gadget("Transpose32", sharing_type.Arithmetic, sharing_type.Arithmetic, "Warren2002", Transpose32S2S2_Warren2002_cost, Transpose32S2S2_Warren2002_memory, "http://www.icodeguru.com/Embedded/Hacker's-Delight/", security_lvl =  None)