# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from itertools import product
import os
import numpy as np
import tomlkit

from . import parameters
from .parameters import *

from . import gadget_class
from .gadget_class import *

###
# some globals
###

# This is for debug only
INTERNAL_VERBOSE=False

# default group resolution file name
default_config_group_file = 'default_config_group_file.toml'
config_group_file_name    = None

# handler of get_cost function (switch between get_cost_value and get_sub_gadgets)
get_cost        = None
# handler of get_memory function (switch between get_memory_value and get_sub_gadgets)
get_memory      = None

list_group_keys = None

# main gadget storage
_gadgets_set  = {}
_group_config = {}


def populate_subgroups(gadget):
	"""
	fill the sub_groups and sub_gadgets dictionary of the input gadget
	To do this automatically, the cost functions are called with a 
	special get_cost and get_memory function.

	Input:
		gadget (gadget_c): a gadget
	"""
	global get_cost, get_memory, list_group_keys
	list_group_keys = []
	# setting get_cost function to an idle one to list the sub-group and sub-gadgets
	get_cost   = globals()['get_sub_gadgets']
	get_memory = globals()['get_sub_gadgets']

	# This is done for vairous values of n.
	# Here is assumed that all sub-gadgets are touched with n in [2, 3, 4, 1<<8]
	for n in [2, 3, 4, 1<<8]:

		scheme_parameters = scheme_related_parameters(12,n)

		try:
			gadget.cost_func(scheme_parameters)
			gadget.memo_func(scheme_parameters)
		except Exception as e:
			print(e)
			print(gadget)
			print(scheme_parameters)
			exit(1)
	
	# back to regular get_cost function
	get_cost   = globals()['get_cost_value']
	get_memory = globals()['get_memory_value']
	for g in list_group_keys:
		group, name = get_group(g)
		if name is None:
			# gadget.sub_groups.add(g)
			gadget.sub_groups+=[g]
		else:
			# gadget.sub_gadgets.add(g)
			gadget.sub_gadgets+=[g]
	list_group_keys = None

def add_new_gadget(gadget_func, input_type, output_type, name, cost_estimation, memory_estimation, ref, security_lvl = None):
	"""
	Adds a new gadget to the gadgets dictionary

	Adds the given input gadget to the dictionary. The name of the gadget will 
	usually contain the initials of the authors and year of publication. 

	Input:
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group
		cost_estimation (function): Function which estimates the cost in number 
		                            of CPU cycle of the gadget
		memory_estimation (function): Function which estimates the memory cost 
		                              of the gadget
		ref (string): Link of the paper from which the gadget is taken 
		security_lvl (security_level): Security level of the gadget
	"""
	gadget_key, in_dict = get_gadget_key(gadget_func, input_type, output_type, name)
	if in_dict:
		print(f"The gadget {gadget_key}  already exists.")
		return False

	gadget = gadget_c(gadget_func, input_type, output_type, name, cost_estimation, memory_estimation, ref, security_lvl)
	populate_subgroups(gadget)

	group  = gadget.group
	if group in _gadgets_set.keys():
		_gadgets_set[group].add_new_gadget(gadget)
	else:
		_gadgets_set[group] = group_c(gadget)
		_group_config[group]= None

	return 0

def get_max_len_group_name():
	return int(max(map(len,_gadgets_set.keys())))

def show_gadgets():
	"""
	Display the whole library of gadgets (in global _gadgets_set). 
	"""

	for gr in _gadgets_set.keys():
		print(_gadgets_set[gr])

	return 0

def gadget_key_in_dict(gadget_key):
	"""
	Tell if the gadget given in input exists in the gadget dictionary

	Input:
		gadget_key (string): Key of the gadget whose existence we want to check 
		                     in the dictionary _gadgets_set

	Output:
		(boolean): True if all the gadgets given in input exist and False if not
	"""

	try:
		group, name = get_group(gadget_key)

		if name is None:
			_ = _gadgets_set[group]
			return True

		return _gadgets_set[group].gadgetExists(name)
	except:
		return False

def get_gadget_key(gadget_func, input_type, output_type, name = None):
	"""
	Give the key of the gadget according to the information provided

	Recomposes the gadget key based on the input information.

	Input:
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group

	Output:
		gadget_key (string): Key of a gadget or a group
		in_dict (Bool): True if the gadget or group exists in _gadgets_set
	"""
	gadget_key = get_key(gadget_func, input_type, output_type, name)

	return gadget_key, gadget_key_in_dict(gadget_key)

def get_security_level(gadget_func, input_type, output_type, name = None):
	"""
	Rerturn the security level of a gadget

	Input:
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group

	Output:
		security_lvl (string): security level of a gadget 
							   if gadget fount in _gadgets_set,
							   None otherwise
	"""
	gadget_key, in_dict = get_gadget_key(gadget_func, input_type, output_type, name)
	
	if not in_dict:
		print(f"The gadget {gadget_key}  don't exist.")
		return None

	group, name = get_group(gadget_key)

	if name is None:
		return None
	
	security_lvl =  _gadgets_set[group].get_gadget(name).security_lvl
	return security_lvl


def get_reference(gadget_func, input_type, output_type, name = None):
	"""
	Rerturn the references information about a gadget

	Input:
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group

	Output:
		reference (string): references information about a gadget 
							if gadget fount in _gadgets_set,
							None otherwise
	"""
	gadget_key, in_dict = get_gadget_key(gadget_func, input_type, output_type, name)
	
	if not in_dict:
		print(f"The gadget {gadget_key}  don't exist.")
		return None

	group, name = get_group(gadget_key)

	if name is None:
		return None
	
	reference =  _gadgets_set[group].get_gadget(name).ref
	return reference

def is_group(gadget_key):
	"""
	Tell if the name given in input is a gadget group

	Takes as input a name in the _gadgets_set dictionary and indicates whether 
	this name corresponds to a group of gadgets or not. A name in the 
	_gadgets_set dictionary corresponding to a group of gadgets is a name that 
	does not contain an underscore, because it just contains the name of the 
	group.

	Input:
		gadget_key (string): A key of the _gadgets_set dictionnary

	Output:
		(boolean): True if name is a group of gadgets and False if not
	"""

	return gadget_key.upper() in _gadgets_set.keys()

def is_good_security_lvl(gadget_func, input_type, output_type, name, required_security_lvl):
	"""
	Indicates whether the input gadget has at least the required security level.

	Input:
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group
		required_security_lvl (security_level): Security level required for the 
		                                        gadget

	Output:
		(boolean): True if name is a group of gadgets and False if not
	"""
	gadget_key, in_dict = get_gadget_key(gadget_func, input_type, output_type, name)
	
	if not in_dict:
		print(f"The gadget {gadget_key}  don't exist.")
		return None

	if is_group(gadget_key):
		return None 

	if required_security_lvl is None:
		return True

	security_lvl = get_security_level(gadget_func, input_type, output_type, name = name)

	if (required_security_lvl == security_level.NI) and isinstance(security_lvl, security_level):
		return True

	if required_security_lvl == security_lvl:
		return True

	return False


def get_sub_gadgets(scheme_parameters, gadget_func, input_type, output_type, name = None, security_lvl = None):
	"""
	This is a get_cost/get_memory function meant to be used to list the sub-groups and sub-gadget call 
	of a gadget. It fills the global list_group_keys list.
	This is used during the initialization of the gadget library (see the add_new_gadget function)

	Input:
		scheme_parameters (scheme_related_parameters): current masking params
		group_func (string): Functionality of gadgets in the group
		input_type (sharing_type): Input type of gadgets in the group
		output_type (sharing_type): Output type of gadgets in the group
		name (string): Gadget name, can be None
		security_lvl (security_level): Security level required for the gadget
	Output:
		cost/memory (int): 0
	"""
	global list_group_keys

	if list_group_keys is None:
		raise NotImplementedError

	gadget_key = get_key(gadget_func, input_type, output_type, name)

	if (gadget_key not in list_group_keys):
		list_group_keys.append(gadget_key)

	return 0


def get_cost_value(scheme_parameters, gadget_func, input_type, output_type, name = None, security_lvl = None):
	"""
	get_cost function: from the paramters get the gadget key. If it is a 
	group, it must be resolved, otherwise the process aborts. Then call the
	performance cost function of the gadget.

	Input:
		scheme_parameters (scheme_related_parameters): current masking params
		group_func (string): Functionality of gadgets in the group
		input_type (sharing_type): Input type of gadgets in the group
		output_type (sharing_type): Output type of gadgets in the group
		name (string): Gadget name, can be None
		security_lvl (security_level): Security level required for the gadget
	Output:
		cost (int): cost in CCE of the gadget given the masking params
	"""
	group_key = get_group_key(gadget_func, input_type, output_type)

	if not group_key in _gadgets_set.keys():
		print(f"The group {group_key}  don't exist.")
		return False

	group = _gadgets_set[group_key]

	if name is None:
		if group.default_gadget is None:
			print(f'Group {group_key} is not set correctly... aborting')
			exit(1)
		if not is_good_security_lvl(group.default_gadget.func, group.default_gadget.input_type, group.default_gadget.output_type, group.default_gadget.name, security_lvl):
			print(f"The default gadget {group.default_gadget.key} doesn't have the good security level.")
			return None

		cost = group.default_gadget.cost_func(scheme_parameters)

	else:
		gadget = group.get_gadget(name)
		if gadget is None:
			print(f'Gadget {name} in group {group_key} is not set correctly... aborting')
			exit(1)

		cost   = gadget.cost_func(scheme_parameters)

	return cost

# By default set get_cost to get_cost_value
get_cost = get_cost_value

def get_memory_value(scheme_parameters, gadget_func, input_type, output_type, name = None, security_lvl = None):
	"""
	get_memory function: from the paramters get the gadget key. If it is a 
	group, it must be resolved, otherwise the process aborts. Then call the
	memory cost function of the gadget.

	Input:
		scheme_parameters (scheme_related_parameters): current masking params
		group_func (string): Functionality of gadgets in the group
		input_type (sharing_type): Input type of gadgets in the group
		output_type (sharing_type): Output type of gadgets in the group
		name (string): Gadget name, can be None
		security_lvl (security_level): Security level required for the gadget
	Output:
		memory (int): memory in bits of the gadget given the masking params
	"""
	group_key = get_group_key(gadget_func, input_type, output_type)

	if not group_key in _gadgets_set.keys():
		print(f"The group {group_key}  don't exist.")
		return False

	group = _gadgets_set[group_key]

	if name is None:
		if group.default_gadget is None:
			print(f'Group {group_key} is not set correctly... aborting')
			exit(1)
		if not is_good_security_lvl(group.default_gadget.func, group.default_gadget.input_type, group.default_gadget.output_type, group.default_gadget.name, security_lvl):
			print(f"The default gadget {group.default_gadget.key} doesn't have the good security level.")
			return None

		memory = group.default_gadget.memo_func(scheme_parameters)

	else:
		gadget = group.get_gadget(name)
		if gadget is None:
			print(f'Gadget {name} in group {group_key} is not set correctly... aborting')
			exit(1)

		memory = gadget.memo_func(scheme_parameters)

	return memory

# By default set get_memory to get_memory_value
get_memory = get_memory_value


def gadgets_in_group(group_func, input_type, output_type):
	"""
	Lists all gadgets belonging to a group

	Builds a list of all keys of gadgets belonging to the group given as input.

	Input:
		group_func (string): Functionality of gadgets in the group
		input_type (sharing_type): Input type of gadgets in the group
		output_type (sharing_type): Output type of gadgets in the group
	Output:
		list_group_gadgets (list of strings): List of all keys of gadgets 
		                                      belonging to the group given as 
		                                      input
	"""
	group_key, in_dict = get_gadget_key(group_func, input_type, output_type)
	if not in_dict:
		print("The group " + group_key + " don't exist.")
		return False

	return [_gadgets_set[group_key].gadgets[name].key for name in _gadgets_set[group_key].gadgets.keys()]


def groups_to_configure(scheme_parameters, group_func, input_type, output_type, name = None, verbose = False):
	"""
	Lists all the groups used by gadgets belonging to a given group

	Builds a list of all the unresolved groups used by gadgets belonging to the input group
	"""

	gadget_key, in_dict = get_gadget_key(group_func, input_type, output_type, name = name)
	if not in_dict:
		print("The group " + gadget_key + " don't exist.")
		return False

	if name is None:
		list_group_gadgets = gadgets_in_group(group_func, input_type, output_type)
	else:
		list_group_gadgets = [gadget_key]

	if verbose:
		print(list_group_gadgets)

	list_groups_to_configure = []
	list_sub_groups          = []

	for gadget_key in list_group_gadgets:
		group, name  = get_group(gadget_key)
		for g in _gadgets_set[group].get_gadget(name).sub_groups:
			if g not in list_sub_groups:
				list_sub_groups += [g]
		for g in _gadgets_set[group].get_gadget(name).sub_gadgets:
			if g not in list_sub_groups:
				list_sub_groups += [g]

	for gadget_key in list_sub_groups:
		if is_group(gadget_key) and (gadget_key not in list_groups_to_configure) and (_group_config[gadget_key] is None):
			list_groups_to_configure += [gadget_key]

	for gadget_key in list_sub_groups:
		subgroup_func, input_type_group, output_type_group, group_name = get_gadget_info(gadget_key)
		list_group_subgadgets = groups_to_configure(scheme_parameters, subgroup_func, input_type_group, output_type_group, group_name)
		for subgroup in list_group_subgadgets:
			if is_group(subgroup) and (subgroup not in list_groups_to_configure) and (_group_config[subgroup] is None):
				list_groups_to_configure += [subgroup]

	if verbose and (len(list_groups_to_configure) > 0):
		print(gadget_key)
		print(list_groups_to_configure)
	return list_groups_to_configure


def gadget_choice_dict(scheme_parameters, group_func, input_type, output_type, name = None):
	"""
	List all available gadgets for each of the unresolved groups found with groups_to_configure
	"""

	gadget_key, in_dict = get_gadget_key(group_func, input_type, output_type, name = name)
	if not in_dict:
		print("The group " + gadget_key + " don't exist.")
		return False

	choice_dict = dict()
	list_groups_to_configure = groups_to_configure(scheme_parameters, group_func, input_type, output_type, name = name, verbose = False)

	for group in list_groups_to_configure:
		subgroup_func, input_type_group, output_type_group, group_name = get_gadget_info(group)
		choice_dict[group]=gadgets_in_group(subgroup_func, input_type_group, output_type_group)
	return choice_dict

def group_config_dict(scheme_parameters, group_func, input_type, output_type):
	"""
	From gadget_choice_dict output, build all possible combination of group resolution and
	sort them by config numbers.
	"""
	choice_dict = gadget_choice_dict(scheme_parameters, group_func, input_type, output_type)
	config_list = list(product(*choice_dict.values()))
	config_dict = dict()
	config_name = "config_"
	i = 1
	for config in config_list:
		config_dict[config_name+str(i)] = dict()
		for gadget in config:
			group, _ = get_group(gadget)
			config_dict[config_name+str(i)][group]=gadget
		i += 1
	return config_dict




def groups_reset():
	"""
	Resets all gadget groups

	Iterate through the entire _gadgets_set dictionary, and for each group, reset 
	its default value to None (i.e. set to unresolved).
	"""
	global _gadgets_set

	for group_key in _gadgets_set.keys():
		_gadgets_set[group_key].default_gadget = None

def set_groups_from_gadgets_list(config, verbose = True):
	"""
	Set default gadgets for each group defined in config

	Input:
		config (dict of strings): each dict key is a group name, pointing to a gadget name
	"""
	global _group_config

	if verbose:
		print(f'Setting unresolved groups:')

	for gadget in config:
		group, name = get_group(gadget)
		_group_config[group] = name 

def set_all_groups_to_defaults():
	"""
	Assigns to each undefined gadget group a default gadget

	Assigns all gadget groups a default gadget (from _group_config).
	"""
	global _gadgets_set

	groups_reset()

	for g in _group_config.keys():
		_gadgets_set[g].default_gadget = _gadgets_set[g].get_gadget(_group_config[g])


def create_group_config_file(config, file_name):
	"""
	Stores the confg inside a file using toml format

	Input:
		config (dict): dictionary where keys are group names and a single 
					   gadget name is associated
		file_name (string): output file name

	"""

	try:
		with open(file_name, mode="wt", encoding="utf-8") as fp:
			tomlkit.dump(config, fp)
	except Exception as e:
		print(e)
		print(f'Error [gadgets_dict.create_group_config_file] group config could not be writen to file {file_name}')

def build_and_set_default_group_configuration_file():
	"""
	Create the default group affectation: for each group, 
	assign the first gadget in the list.
	Create the group config file at the same time
	"""
	global config_group_file_name, _group_config
	
	config = {}

	for g in _gadgets_set.keys():
		# select the first gadget that match the group
		name = list(_gadgets_set[g].gadgets.keys())[0]
		config[g] = g+'_'+name
	
		_group_config[g] = config[g]

	config_group_file_name = default_config_group_file

	create_group_config_file(config, default_config_group_file)

def update_group_defaults(config_file = None, verbose = True):
	"""
	Stores the confg inside a file using toml format

	Input:
		config_file (name): file name (can be None) 
							where a group configuration 
							is stored

	"""

	global config_group_file_name, _group_config

	groups_reset()

	if config_file is None:
		# nothing to be done, just keep all entries of _group_config to None
		return True

	success = True
	try:
		with open(config_file, mode="rt", encoding="utf-8") as fp:
			config = tomlkit.load(fp)

		for g in config.keys():
			_  = _group_config[g]
			if config[g].upper() == "ALL":
				_group_config[g] = None
			else:
				_group_config[g] = config[g].upper()

		config_group_file_name = config_file

	except Exception as e:
		print(e)
		print(f'Error [gadgets_dict.update_group_defaults]: config file does not set all groups correctly... groups will be set to default gadgets (first appearing in the gadget dict, i.e. quite arbitrary). Please have a look to the default configuration file: {default_config_group_file}')
		build_and_set_default_group_configuration_file()
		success = False

	return success


