# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from copy import deepcopy
import tomlkit

"""
	File that allows to create classes called scheme_related_parameters. These 
	classes make it possible to define the masking parameters of the 
	cryptographic scheme on which we want to place ourselves to make our 
	estimates. These parameters are:
	- the number of shares for a masked value
	- the size of the shares of our masked values
	These classes also contain methods to modify the various parameters.
"""

class scheme_related_parameters:
	"""
		Class which allows to define the masking parameters of a cryptographic 
		scheme.
	"""

	def __init__(self, params_config_file, n=-1):
		"""
			Scheme parameter initialization

			n (int): Number of shares of masked values. (default -1 is an incorrect value for n, should be >= 2)
			bitsize_shares (int): Bit size of shares. It depends on cryptographic primitive / masking scheme
		"""
		try:
			self.n = int(n)
		except:
			print(f'Error [scheme_related_parameters.__init__]: paramter {n} could be cast to an integer!')
			print(f'aborting')
			exit(1)

		try: 
			# first try to read the input as an int
			self.bitsize_shares = int(params_config_file)
		except:
			try:
				with open(params_config_file, mode="rt", encoding="utf-8") as fp:
					config = tomlkit.load(fp)

				self.bitsize_shares = config['bitsize_shares']
			except:
				print(f'Error [scheme_related_parameters.__init__]: paramter configuration file {params_config_file} could be read properly!')
				print(f'Default kyber parameters will be used instead')
					
				self.bitsize_shares = 12

	def set_bitsize_shares(self, new_bitsize_shares):
		"""
			Method that allows to change the bit size of shares

			Input:
				new_bitsize_shares (int): New bit size of shares
		"""
		self.bitsize_shares = new_bitsize_shares

	def new_number_of_shares(self, new_n):
		"""
			Method that allows to change the number of shares

			Input:
				new_n (int): New number of shares
		"""
		self.n = new_n

	def __str__(self):
		out =  f'Scheme Parameter:\n'
		out += f'\t{self.n=}\n'
		out += f'\t{self.bitsize_shares=}\n'

		return out



