# Inspector Gadget
# Copyright (C) 2023  Camille Mutschler <camille.mutschler@lirmm.fr>
#                     Thomas Roche <thomas@ninjalab.io>
#                     Laurent Imbert <laurent.imbert@lirmm.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os, sys
from itertools import product
from copy import deepcopy
from tqdm.auto import tqdm

if __name__ == "__main__":
	from gogogadgets import gadgets_dict
	from gogogadgets.parameters import *
	from gogogadgets.gadgets_dict import sharing_type
	from gogogadgets import atomic_operations
else:
	from .gogogadgets import gadgets_dict
	from .gogogadgets.parameters import *
	from .gogogadgets.gadgets_dict import sharing_type
	from .gogogadgets import atomic_operations

def gadget_estimation(params_config_file, gadget_func, input_type, output_type, name = None, n=2, verbose = True):
	"""
	Generates the performance and memory cost estimations of a given gadget
	Input gadget should exist in gadgets_dict._gadget_set and not be a group.
	All groups are assumed to be resolved.

	Input:
		params_config_file (string): Scheme parameters file
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group
		n (int): Number of shares of masked values
		verbose (bool): Display comments or not
	"""

	# store register/memory space for evaluation
	register_space  = atomic_operations.hardware_model['register_space']
	memory_space    = atomic_operations.hardware_model['memory_space']

	try:
		n = int(n)
		assert(n > 1)
	except:
		print(f"Error: the number of shares n={n} cannot be cast to an int greater or equal to 2.")
		return False
	
	gadget_key, in_dict = gadgets_dict.get_gadget_key(gadget_func, input_type, output_type, name)
	if not in_dict:
		print("Error: The gadget " + gadget_key + " don't exist.")
		return False

	scheme_parameters_cost   = scheme_related_parameters(params_config_file, n)
	scheme_parameters_memory = deepcopy(scheme_parameters_cost)

	try:
		memory = gadgets_dict.get_memory(scheme_parameters_memory, gadget_func, input_type, output_type, name)
	except TypeError as e:
		# print(e)
		return None, None, False

	if memory is None:
		return None, None, False

	try:
		memory_space = int(memory_space)
		if memory > memory_space:
			if verbose:
				print("The computation is too expensive in memory to run.")
			return None, memory, False
	except:
		if memory_space != "infinity":
			raise NotImplementedError

	patched_load_store_cost  = False
	fit_in_registers         = True

	try:
		register_space = int(register_space)
		if memory < register_space:
			if verbose:
				print("The estimated CPU cycle cost is based on the assumption that all values are stored in registers.")
			atomic_operations.atomic_operations["load_op"]["cost"] = 0
			atomic_operations.atomic_operations["store_op"]["cost"] = 0
			patched_load_store_cost = True
		else:
			if verbose:
				print("The estimated CPU cycle cost is based on the assumption that the potential tables used are stored in memory, as the amount of data is too large to fit in the registers space.")
			fit_in_registers = False
	except:
		if register_space == "infinity":
			if verbose:
				print("The estimated CPU cycle cost is based on the assumption that all values are stored in registers.")
			atomic_operations.atomic_operations["load_op"]["cost"] = 0
			atomic_operations.atomic_operations["store_op"]["cost"] = 0
			patched_load_store_cost = True
		else:
			raise NotImplementedError

	try:
		cost = gadgets_dict.get_cost(scheme_parameters_cost, gadget_func, input_type, output_type, name)
	except TypeError as e:
		# print(e)
		return None, None, False

	if patched_load_store_cost:
		atomic_operations.atomic_operations["load_op"]["cost"]  = atomic_operations.atomic_operations["LDR"]["cost"]
		atomic_operations.atomic_operations["store_op"]["cost"] = atomic_operations.atomic_operations["STR"]["cost"]

	if verbose:
		print("\n")
	return cost, memory, fit_in_registers

def kyber_gadget_estimation(gadget_func, input_type, output_type, name = None, n=2, verbose = True):
	cost, memory, fit_in_registers = gadget_estimation("examples_masking_config/kyber_params.toml", gadget_func, input_type, output_type, name = name, n=n, verbose = verbose)
	return cost, memory, fit_in_registers

def gadget_estimation_single_group_config(params_config_file, gadget_func, input_type, output_type, name, list_n=[2], file = None, verbose = True):
	"""
	Generates the performance and memory cost estimations of a given gadget for a list of number of shares.
	If input gadget is a group, the function will iterate over all gadgets in the group.
	All groups are assumed to be resolved.

	Input:
		params_config_file (string): Scheme parameters file
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group
		list_n (List of int): Number of shares of masked values
		file: open file for writing output results (or None)
		verbose (bool): Display comments or not
	"""

	gadgets_dict.set_all_groups_to_defaults()
	gadget_key, in_dict = gadgets_dict.get_gadget_key(gadget_func, input_type, output_type, name)

	if name is None:
		list_primary_gadgets = gadgets_dict.gadgets_in_group(gadget_func, input_type, output_type)
		file.write(f"The gadget {gadget_key} is a group, all {len(list_primary_gadgets)} gadgets appearing in the group will be estimated\n")
	else:
		list_primary_gadgets = [gadget_key]

	for gadget_key in list_primary_gadgets:

		if name is None:
			file.write(f'Gadget: {gadget_key}\n')

		gadget_func, input_type, output_type, gadget_name = gadgets_dict.get_gadget_info(gadget_key)

		for n in list_n:

			if file is not None:
				file.write(f'n={n}\n')

			cost, memory, fit_in_registers = gadget_estimation(params_config_file, gadget_func, input_type, output_type, name = gadget_name, n=n, verbose = verbose)

			if file is not None:
				file.write(f'Cost: {cost} CCE\n')
				file.write(f'Memory: {memory} bits\n')
				file.write(f'Fit in registers: {fit_in_registers}\n\n')
		

def gadget_estimation_main(params_config_file, gadget_func, input_type, output_type, name, list_n=[2], filename="test", result_folder = "../Estimation_Results", verbose = True):
	"""
	Generates the performance and memory cost estimations of a given gadget for a list of number of shares.
	If input gadget is a group, the function will iterate over all gadgets in the group.
	For all groups that are not resolved yet, iterate over all gadgets in the group.

	Input:
		params_config_file (string): Scheme parameters file
		gadget_func (string): Functionality of the gadget
		input_type (sharing_type): Input type of the gadget
		output_type (sharing_type): Output type of the gadget
		name (string): Name of this gadget from the given group
		list_n (List of int): Number of shares of masked values
		filename (string): name of the file where perf and memory results are stored
		result_folder (string): path to the where the result file is writen
		verbose (bool): Display comments or not
	"""

	run_on_a_group = False

	if atomic_operations.hardware_model is None:
		print(f"The hardware model is not yet set... aborting")
		return False

	gadget_key, in_dict = gadgets_dict.get_gadget_key(gadget_func, input_type, output_type, name)

	if not in_dict:
		print(f"The gadget {gadget_key}  don't exist.")
		return False

	if filename is not None:
		if not os.path.exists(result_folder):
			os.makedirs(result_folder)

		estimations_file = result_folder+"/results_"+filename+".txt"
		file = open(estimations_file, mode="w", encoding="utf-8")
	else:
		file = sys.stdout

	file.write('\n')
	file.write('Configuration files:\n')
	file.write(f'\tcryptographic scheme parameters: {params_config_file} \n')
	file.write(f'\thardware model parameters      : {atomic_operations.config_hardware_file_name} \n')
	file.write(f'\tdefault group resolution       : {gadgets_dict.config_group_file_name} \n')
	file.write('\n')
	file.write(f'Estimation of the {"group" if name is None else "gadget"} {gadget_key}\n')
	file.write(f'\tGadget functionality: {gadget_func}\n')
	file.write(f'\tInput type          : {input_type}\n')
	file.write(f'\tOutput type         : {output_type}\n')
	file.write(f'\tName                : {name}\n')
	if name is not None:
		file.write(f'\tReference           : {gadgets_dict.get_reference(gadget_func, input_type, output_type, name)}\n')
		file.write(f'\tSecurity level      : {gadgets_dict.get_security_level(gadget_func, input_type, output_type, name)}\n\n')

	scheme_parameters     = scheme_related_parameters(params_config_file)
	list_undefined_groups = gadgets_dict.gadget_choice_dict(scheme_parameters, gadget_func, input_type, output_type, name)

	if len(list_undefined_groups) == 0:
		gadget_estimation_single_group_config(params_config_file, gadget_func, input_type, output_type, name = name, list_n=list_n, file = file, verbose = verbose)

	else:
		config_list = list(product(*list_undefined_groups.values()))

		file.write(f'Some groups were not resoved by the Group Parameter File {gadgets_dict.config_group_file_name}\n')
		file.write(f'Here is the list of unresolved groups and corresponding gadgets:\n')
		for g in list_undefined_groups.keys():
			file.write(f'Group {g}:\n')
			for gad in list_undefined_groups[g]:
				file.write(f'\t{gad}\n')
		file.write(f'Each of the {len(config_list)} combinations will be tested\n\n')

		for config in config_list:
			file.write(f'Setting unresolved groups:\n')
			for gadget in config:
				group, _ = gadgets_dict.get_group(gadget)
				file.write(f'\t{group:{gadgets_dict.get_max_len_group_name()}s} : {gadget}\n')
			file.write(f'\n')

			gadgets_dict.set_groups_from_gadgets_list(config, verbose = False)
			gadget_estimation_single_group_config(params_config_file, gadget_func, input_type, output_type, name = name, list_n=list_n, file = file, verbose = verbose)

	file.write('# End of the script\n')

	if filename is not None:
		file.close()
	

if __name__ == "__main__":

	"""
		Main

		if group_config_file is not set, then all groups are considered unresolved and all possible combinations of 
		sub-gadgets will be tested iteratively... this grows large fast.
	"""

	parser = argparse.ArgumentParser()
	parser.add_argument("--component_config_file", type=str, default="examples_hardware_config/Cortex_M3_rand=32.toml", help="hardware component configuration file")
	parser.add_argument("--params_config_file",    type=str, default="examples_masking_config/kyber_params.toml", help="scheme parameters file")
	parser.add_argument("--group_config_file",     type=str, default=None, help="group configuration file")
	parser.add_argument("-gf", "--gadget_func",    type=str, default="KyberCompress256coeffs", help="gadget_functionnality")
	parser.add_argument("-in", "--input_type",     type=str, choices=sharing_type.__members__.keys(), default="Arithmetic", help="input type")
	parser.add_argument("-out", "--output_type",   type=str, choices=sharing_type.__members__.keys(), default="Boolean", help="output type")
	parser.add_argument("-n", "--gadget_name",     type=str, default=None, help="gadget name")
	parser.add_argument("-ln", "--list_n",         type=str, default='[2]', help="list of number of shares to test")
	parser.add_argument("--result_name",           type=str, default=None, help="result file name will be 'result_{args.result_name}.txt', if None then result is sent to stdout")
	parser.add_argument("--result_file_path",      type=str, default='../', help="where to put the result file, if any")
	parser.add_argument("--create_default_group_conf", action='store_true', help="create the default group configuration file default_config_group_file.toml. disregard all other options")
	parser.add_argument("--display_gadgets",           action='store_true', help="display all available gadgets. disregard all other options")
	args = parser.parse_args()

	if args.create_default_group_conf:
		gadgets_dict.build_and_set_default_group_configuration_file()
		exit(0)

	if args.display_gadgets:
		gadgets_dict.show_gadgets()
		exit(0)
	
	atomic_operations.set_atomic_operations_costs(args.component_config_file)
	gadgets_dict.update_group_defaults(args.group_config_file)

	params_config_file = args.params_config_file
	gadget_func        = args.gadget_func
	input_type         = getattr(sharing_type, args.input_type)
	output_type        = getattr(sharing_type, args.output_type)
	gadget_name        = args.gadget_name
	list_n             = eval(args.list_n)
	result_name        = args.result_name
	result_file_path   = args.result_file_path

	gadget_estimation_main(params_config_file, gadget_func, input_type, output_type, name=gadget_name, list_n=list_n, filename = result_name, result_folder = result_file_path, verbose = False)


	
